import asyncio
import multiprocessing
import threading

from dataclasses import dataclass
from typing import Generator

import pytest

from hololinked.core.exceptions import BreakLoop
from hololinked.core.zmq.brokers import (
    AsyncZMQClient,
    AsyncZMQServer,
    MessageMappedZMQClientPool,
    SyncZMQClient,
)
from hololinked.core.zmq.message import (
    ERROR,
    EXIT,
    HANDSHAKE,
    INVALID_MESSAGE,
    REPLY,
    TIMEOUT,
    RequestMessage,
    ResponseMessage,
    SerializableData,
)
from hololinked.utils import get_current_async_loop, uuid_hex


try:
    from .conftest import AppIDs as MessageAppIDs
    from .test_01_message import validate_response_message
except ImportError:
    from conftest import AppIDs as MessageAppIDs
    from test_01_message import validate_response_message


@dataclass
class AppIDs:
    """
    Application related IDs generally used by end-user,
    like server, client, and thing IDs.
    """

    server_id: str
    """RPC server ID"""
    thing_id: str
    """A thing ID"""
    sync_client_id: str
    """A synchronous client ID"""
    async_client_id: str
    """An asynchronous client ID"""
    msg_mapped_async_client_id: str
    """A message-mapped asynchronous client ID"""


@pytest.fixture(scope="module")
def app_ids() -> AppIDs:
    """Generate unique test IDs for server, client, and thing for each test"""
    return AppIDs(
        server_id=f"test-server-{uuid_hex()}",
        thing_id=f"test-thing-{uuid_hex()}",
        sync_client_id=f"test-sync-client-{uuid_hex()}",
        async_client_id=f"test-async-client-{uuid_hex()}",
        msg_mapped_async_client_id=f"test-mapped-async-client-{uuid_hex()}",
    )


@pytest.fixture(scope="module")
def server(app_ids: AppIDs) -> Generator[AsyncZMQServer, None, None]:
    server = AsyncZMQServer(id=app_ids.server_id)
    yield server
    # exit written in thread
    # server.exit()


@pytest.fixture(scope="module")
def sync_client(app_ids: AppIDs) -> Generator[SyncZMQClient, None, None]:
    client = SyncZMQClient(id=app_ids.sync_client_id, server_id=app_ids.server_id, handshake=False)
    yield client
    client.exit()


@pytest.fixture(scope="module")
def async_client(app_ids: AppIDs) -> Generator[AsyncZMQClient, None, None]:
    client = AsyncZMQClient(id=app_ids.async_client_id, server_id=app_ids.server_id, handshake=False)
    yield client
    client.exit()


@pytest.fixture(scope="module")
def message_mapped_client(app_ids: AppIDs) -> Generator[MessageMappedZMQClientPool, None, None]:
    client = MessageMappedZMQClientPool(
        id="client-pool",
        client_ids=[app_ids.msg_mapped_async_client_id],
        server_ids=[app_ids.server_id],
        handshake=False,
    )
    client._client_to_thing_map[app_ids.msg_mapped_async_client_id] = app_ids.thing_id
    client._thing_to_client_map[app_ids.thing_id] = app_ids.msg_mapped_async_client_id
    yield client
    client.exit()


def run_zmq_server(server: AsyncZMQServer, done_queue: multiprocessing.Queue) -> None:
    event_loop = get_current_async_loop()

    async def run():
        while True:
            try:
                messages = await server.async_recv_requests()
                for message in messages:
                    if message.type == EXIT:
                        server.exit()
                        break
                await asyncio.sleep(0.01)
            except BreakLoop:
                break

    event_loop.run_until_complete(run())
    event_loop.run_until_complete(asyncio.gather(*asyncio.all_tasks(event_loop)))
    if done_queue:
        done_queue.put(True)


@pytest.fixture(scope="module", autouse=True)
def start_server(server: AsyncZMQServer, sync_client: SyncZMQClient, app_ids: AppIDs):
    done_queue = multiprocessing.Queue()
    thread = threading.Thread(target=run_zmq_server, args=(server, done_queue), daemon=True)
    thread.start()
    yield thread
    request_message = RequestMessage.craft_with_message_type(
        receiver_id=app_ids.server_id,
        sender_id=app_ids.sync_client_id,
        message_type=EXIT,
    )
    sync_client.socket.send_multipart(request_message.byte_array)
    done = done_queue.get(timeout=3)
    if done:
        thread.join()
    else:
        print("Server did not properly process exit request")


def test_01_01_sync_client_handshake_complete(sync_client: SyncZMQClient):
    sync_client.handshake()
    assert sync_client._monitor_socket is not None
    assert sync_client._monitor_socket in sync_client.poller


async def test_01_02_sync_client_basic_message_contract_types(
    sync_client: SyncZMQClient,
    server: AsyncZMQServer,
    app_ids: AppIDs,
) -> None:
    active_app_ids = MessageAppIDs(
        server_id=app_ids.server_id, thing_id=app_ids.thing_id, client_id=app_ids.sync_client_id
    )
    request_message = RequestMessage.craft_from_arguments(
        receiver_id=app_ids.server_id,
        sender_id=app_ids.sync_client_id,
        thing_id=app_ids.thing_id,
        objekt="some_prop",
        operation="readproperty",
    )

    await server._handle_timeout(request_message, timeout_type="execution")
    await server._handle_invalid_message(request_message, SerializableData(Exception("test")))
    await server._handshake(request_message)
    await server._handle_error_message(request_message, Exception("test"))
    await server.async_send_response(request_message)
    await server.async_send_response_with_message_type(request_message, ERROR, SerializableData(Exception("test")))

    msg = sync_client.recv_response(request_message.id)
    assert msg.type == TIMEOUT
    validate_response_message(msg, app_ids=active_app_ids)

    msg = sync_client.recv_response(request_message.id)
    assert msg.type == INVALID_MESSAGE
    validate_response_message(msg, app_ids=active_app_ids)

    msg = sync_client.socket.recv_multipart()
    response_message = ResponseMessage(msg)
    assert response_message.type == HANDSHAKE
    validate_response_message(response_message, app_ids=active_app_ids)

    msg = sync_client.recv_response(request_message.id)
    assert msg.type == ERROR
    validate_response_message(msg, app_ids=active_app_ids)

    msg = sync_client.recv_response(request_message.id)
    assert msg.type == REPLY
    validate_response_message(msg, app_ids=active_app_ids)

    msg = sync_client.recv_response(request_message.id)
    assert msg.type == ERROR
    validate_response_message(msg, app_ids=active_app_ids)
    sync_client.handshake()


async def test_01_03_sync_client_polling(sync_client: SyncZMQClient, server: AsyncZMQServer):
    done = asyncio.Future()

    async def verify_poll_stopped():
        await server.poll_requests()
        server.poll_timeout = 1000
        await server.poll_requests()
        done.set_result(True)

    async def stop_poll():
        await asyncio.sleep(0.1)
        server.stop_polling()
        await asyncio.sleep(0.1)
        server.stop_polling()

    await asyncio.gather(verify_poll_stopped(), stop_poll())
    await done
    assert server.poll_timeout == 1000
    sync_client.handshake()


async def test_async_client_handshake_complete(async_client: AsyncZMQClient):
    async_client.handshake()
    await async_client.handshake_complete()
    assert async_client._monitor_socket is not None
    assert async_client._monitor_socket in async_client.poller


async def test_02_01_async_client_message_contract_types(
    async_client: AsyncZMQClient,
    server: AsyncZMQServer,
    app_ids: AppIDs,
) -> None:
    active_app_ids = MessageAppIDs(
        server_id=app_ids.server_id,
        thing_id=app_ids.thing_id,
        client_id=app_ids.async_client_id,
    )

    request_message = RequestMessage.craft_from_arguments(
        receiver_id=app_ids.server_id,
        sender_id=app_ids.async_client_id,
        thing_id=app_ids.thing_id,
        objekt="some_prop",
        operation="readproperty",
    )

    await server._handle_timeout(request_message, timeout_type="invokation")
    await server._handle_invalid_message(request_message, SerializableData(Exception("test1")))
    await server._handshake(request_message)
    await server._handle_error_message(request_message, Exception("test2"))
    await server.async_send_response(request_message)
    await server.async_send_response_with_message_type(request_message, ERROR, SerializableData(Exception("test3")))

    msg = await async_client.async_recv_response(request_message.id)
    assert msg.type == TIMEOUT
    validate_response_message(msg, app_ids=active_app_ids)

    msg = await async_client.async_recv_response(request_message.id)
    assert msg.type == INVALID_MESSAGE
    validate_response_message(msg, app_ids=active_app_ids)

    msg = await async_client.socket.recv_multipart()
    response_message = ResponseMessage(msg)
    assert response_message.type == HANDSHAKE
    validate_response_message(response_message, app_ids=active_app_ids)

    msg = await async_client.async_recv_response(request_message.id)
    assert msg.type == ERROR
    validate_response_message(msg, app_ids=active_app_ids)

    msg = await async_client.async_recv_response(request_message.id)
    assert msg.type == REPLY
    validate_response_message(msg, app_ids=active_app_ids)

    msg = await async_client.async_recv_response(request_message.id)
    assert msg.type == ERROR
    validate_response_message(msg, app_ids=active_app_ids)


async def test_03_01_mapped_handshake_complete(message_mapped_client: MessageMappedZMQClientPool):
    message_mapped_client.handshake()
    await message_mapped_client.handshake_complete()
    for client in message_mapped_client.pool.values():
        assert client._monitor_socket is not None
        assert client._monitor_socket in message_mapped_client.poller


async def test_mapped_message_contract_types(
    message_mapped_client: MessageMappedZMQClientPool,
    server: AsyncZMQServer,
    app_ids: AppIDs,
) -> None:
    active_app_ids = MessageAppIDs(
        server_id=app_ids.server_id,
        thing_id=app_ids.thing_id,
        client_id=app_ids.msg_mapped_async_client_id,
    )
    request_message = RequestMessage.craft_from_arguments(
        receiver_id=app_ids.server_id,
        sender_id=app_ids.msg_mapped_async_client_id,
        thing_id=app_ids.thing_id,
        objekt="some_prop",
        operation="readproperty",
    )

    message_mapped_client.start_polling()

    message_mapped_client.events_map[request_message.id] = message_mapped_client.event_pool.pop()
    await server._handle_timeout(request_message, timeout_type="invokation")
    msg = await message_mapped_client.async_recv_response(app_ids.thing_id, request_message.id)
    assert msg.type == TIMEOUT
    validate_response_message(msg, app_ids=active_app_ids)

    message_mapped_client.events_map[request_message.id] = message_mapped_client.event_pool.pop()
    await server._handle_invalid_message(request_message, SerializableData(Exception("test")))
    msg = await message_mapped_client.async_recv_response(app_ids.thing_id, request_message.id)
    assert msg.type == INVALID_MESSAGE
    validate_response_message(msg, app_ids=active_app_ids)

    message_mapped_client.events_map[request_message.id] = message_mapped_client.event_pool.pop()
    await server._handshake(request_message)
    msg = await message_mapped_client.pool[app_ids.msg_mapped_async_client_id].socket.recv_multipart()
    response_message = ResponseMessage(msg)
    assert response_message.type == HANDSHAKE
    validate_response_message(response_message, app_ids=active_app_ids)

    message_mapped_client.events_map[request_message.id] = message_mapped_client.event_pool.pop()
    await server.async_send_response(request_message)
    msg = await message_mapped_client.async_recv_response(app_ids.thing_id, request_message.id)
    assert msg.type == REPLY
    validate_response_message(msg, app_ids=active_app_ids)

    message_mapped_client.events_map[request_message.id] = message_mapped_client.event_pool.pop()
    await server.async_send_response_with_message_type(request_message, ERROR, SerializableData(Exception("test")))
    msg = await message_mapped_client.async_recv_response(app_ids.thing_id, request_message.id)
    assert msg.type == ERROR
    validate_response_message(msg, app_ids=active_app_ids)

    message_mapped_client.stop_polling()


async def test_03_02_mapped_verify_polling(message_mapped_client: MessageMappedZMQClientPool):
    done = asyncio.Future()

    async def verify_poll_stopped():
        await message_mapped_client.poll_responses()
        message_mapped_client.poll_timeout = 1000
        await message_mapped_client.poll_responses()
        done.set_result(True)

    async def stop_poll():
        await asyncio.sleep(0.1)
        message_mapped_client.stop_polling()
        await asyncio.sleep(0.1)
        message_mapped_client.stop_polling()

    await asyncio.gather(verify_poll_stopped(), stop_poll())
    await done
    assert message_mapped_client.poll_timeout == 1000


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



from .server import BaseProtocolServer, run, stop  # noqa: F401, isort: skip
from .http import HTTPServer  # noqa: F401
from .mqtt import MQTTPublisher  # noqa: F401
from .zmq import ZMQServer  # noqa: F401


from .security import (  # noqa: F401, isort: skip
    APIKeySecurity,
    Argon2BasicSecurity,
    BcryptBasicSecurity,
    OIDCSecurity,
)



from typing import Optional


def get_summary(docs: str) -> Optional[str]:
    """
    Return the first line of the dosctring of an object

    Parameters
    ----------
    docs:
        The docstring of the object

    Returns
    -------
    str:
        First line of object docstring
    """
    if docs:
        return docs.partition("\n")[0].strip()
    else:
        return ""



"""
adapted from pyro - https://github.com/irmen/Pyro5 - see following license

MIT License

Copyright (c) Irmen de Jong

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
import os
import shutil
import tracemalloc
import warnings

from typing import Any  # noqa: F401

import zmq.asyncio

from .utils import generate_main_script_log_filename, set_global_event_loop_policy


class Configuration:
    """
    Allows to auto apply common settings used throughout the package, instead of passing these settings as arguments.
    Import `global_config` variable instead of instantiating this class. Please check `global_config` docstring for supported values
    or [website documentation](https://docs.hololinked.dev/api-reference/global-config/).

    This implementation needs to be improved in general. Consider opening an issue
    if you have suggestions at [GitHub](https://github.com/hololinked-dev/hololinked/issues).
    """

    __slots__ = [
        "app_name",
        # folders
        "TEMP_DIR",
        # TCP sockets
        "TCP_SOCKET_SEARCH_START_PORT",
        "TCP_SOCKET_SEARCH_END_PORT",
        # HTTP server
        "ALLOW_CORS",
        # database
        "DB_CONFIG_FILE",
        # Eventloop
        "USE_UVLOOP",
        "TRACE_MALLOC",
        # schema validation
        "VALIDATE_SCHEMAS",
        # ZMQ
        "ZMQ_CONTEXT",
        # make debugging easier
        "DEBUG",
        # logging
        "LOG_LEVEL",
        "USE_LOG_FILE",
        "LOG_FILENAME",
        "ROTATE_LOG_FILES",
        "LOGFILE_BACKUP_COUNT",
        # "USE_STRUCTLOG",
        "COLORED_LOGS",
        # serializers
        "ALLOW_PICKLE",
        "ALLOW_UNKNOWN_SERIALIZATION",
        # internal
        "_sockets_folder",
        "_secrets_folder",
        "_logs_folder",
        "_db_folder",
    ]

    def __init__(self, app_name: str | None = None):
        self.app_name = app_name
        self._sockets_folder = "sockets"
        self._secrets_folder = "secrets"
        self._logs_folder = "logs"
        self._db_folder = "db"
        self.load_variables()

    def load_variables(self):
        """Set default values. This method is called during `__init__`"""
        # note that all variables have not been implemented yet,
        # things just come and go as of now
        self.TEMP_DIR = os.path.join(os.path.expanduser("~"), ".hololinked")
        self.TCP_SOCKET_SEARCH_START_PORT = 60000
        self.TCP_SOCKET_SEARCH_END_PORT = 65535
        self.ALLOW_CORS = False
        self.DB_CONFIG_FILE = None
        self.USE_UVLOOP = False
        self.TRACE_MALLOC = False
        # self.VALIDATE_SCHEMA_ON_CLIENT = False
        self.VALIDATE_SCHEMAS = True
        self.ZMQ_CONTEXT = zmq.asyncio.Context()
        self.DEBUG = False
        self.LOG_LEVEL = logging.DEBUG if self.DEBUG else logging.INFO
        # self.USE_STRUCTLOG = True
        self.COLORED_LOGS = False
        self.USE_LOG_FILE = False
        self.LOG_FILENAME = os.path.join(self.TEMP_DIR_LOGS, generate_main_script_log_filename(self.app_name))
        self.ROTATE_LOG_FILES = True
        self.LOGFILE_BACKUP_COUNT = 14
        # Add the filename of the main script importing this module
        self.ALLOW_PICKLE = False
        self.ALLOW_UNKNOWN_SERIALIZATION = False

        self.setup()

    def setup(self):
        """
        Actions to be done to recreate global configuration state after changing config values.
        Called after `load_variables` and `set` methods.

        Please call this method after changing config values directly specific to logging or event loop policy
        """
        self.setup_temp_dirs()

        set_global_event_loop_policy(self.USE_UVLOOP)
        if self.TRACE_MALLOC:
            tracemalloc.start()

        from .logger import setup_logging

        self.LOG_LEVEL = logging.DEBUG if self.DEBUG else self.LOG_LEVEL

        # if self.USE_STRUCTLOG: # no other option for now
        setup_logging(
            log_level=self.LOG_LEVEL,
            colored_logs=self.COLORED_LOGS,
            log_file=self.LOG_FILENAME if self.USE_LOG_FILE else None,
            rotate_log_files=self.ROTATE_LOG_FILES,
            logfile_backup_count=self.LOGFILE_BACKUP_COUNT,
        )

    def copy(self):
        """returns a copy of this config as another object"""
        other = object.__new__(Configuration)
        for item in self.__slots__:
            setattr(other, item, getattr(self, item))
        return other

    def set(self, **kwargs):
        """
        sets multiple config values at once, and recreates necessary global states.
        `load_variables` sets default values first, then overwrites with environment file values.
        This method only overwrites the specified values.
        """
        for item, value in kwargs.items():
            setattr(self, item, value)
        self.setup()

    def asdict(self):
        """returns this config as a regular dictionary"""
        return {item: getattr(self, item) for item in self.__slots__}

    def zmq_context(self) -> zmq.asyncio.Context:
        """
        Returns a global ZMQ async context. Use socket_class argument to retrieve
        a synchronous socket if necessary.
        """
        return self.ZMQ_CONTEXT

    def set_default_server_execution_context(
        self,
        invokation_timeout: int | None = None,
        execution_timeout: int | None = None,
        oneway: bool = False,
    ) -> None:
        """Sets the default server execution context for the application"""
        from .core.zmq.message import default_server_execution_context

        default_server_execution_context.invokationTimeout = invokation_timeout or 5
        default_server_execution_context.executionTimeout = execution_timeout or 5
        default_server_execution_context.oneway = oneway

    def set_default_thing_execution_context(
        self,
        fetch_execution_logs: bool = False,
    ) -> None:
        """Sets the default thing execution context for the application"""
        from .core.zmq.message import default_thing_execution_context

        default_thing_execution_context.fetchExecutionLogs = fetch_execution_logs

    @property
    def TEMP_DIR_SOCKETS(self) -> str:
        """returns the temporary directory path for IPC sockets"""
        return os.path.join(self.TEMP_DIR, self._sockets_folder)

    @property
    def TEMP_DIR_LOGS(self) -> str:
        """returns the temporary directory path for log files"""
        return os.path.join(self.TEMP_DIR, self._logs_folder)

    @property
    def TEMP_DIR_DB(self) -> str:
        """returns the temporary directory path for database files"""
        return os.path.join(self.TEMP_DIR, self._db_folder)

    @property
    def TEMP_DIR_SECRETS(self) -> str:
        """returns the temporary directory path for secret files"""
        return os.path.join(self.TEMP_DIR, self._secrets_folder)

    def setup_temp_dirs(self) -> None:
        for directory in [
            self.TEMP_DIR,
            self.TEMP_DIR_SOCKETS,
            self.TEMP_DIR_LOGS,
            self.TEMP_DIR_DB,
            self.TEMP_DIR_SECRETS,
        ]:
            try:
                os.mkdir(directory)
            except FileExistsError:
                pass
            except PermissionError:
                warnings.warn(f"permission denied to create directory {directory}", UserWarning)

    def set_temp_dir(self, path: str) -> None:
        """sets the base directory path for temporary files and application data (sockets, logs, databases, secrets)"""
        self.TEMP_DIR = path
        self.setup_temp_dirs()

    def set_sockets_folders(self, path: str) -> None:
        """sets the temporary directory path for IPC sockets"""
        self._sockets_folder = path
        self.setup_temp_dirs()

    def set_logs_folder(self, path: str) -> None:
        """sets the temporary directory path for log files"""
        self._logs_folder = path
        self.setup_temp_dirs()

    def set_db_folder(self, path: str) -> None:
        """sets the temporary directory path for database files"""
        self._db_folder = path
        self.setup_temp_dirs()

    def set_secrets_folder(self, path: str) -> None:
        """sets the temporary directory path for secret files"""
        self._secrets_folder = path
        self.setup_temp_dirs()

    def cleanup_temp_dirs(self, cleanup_databases: bool = False) -> None:
        """
        Cleans up temporary directories used by hololinked, all log files and IPC sockets are removed.
        If `cleanup_databases` is `True`, database files are also removed.
        """
        directories = [self.TEMP_DIR_SOCKETS, self.TEMP_DIR_LOGS]
        if cleanup_databases:
            directories.append(self.TEMP_DIR_DB)
        for directory in directories:
            try:
                shutil.rmtree(directory)
            except FileNotFoundError:
                pass
            except PermissionError:
                warnings.warn(f"permission denied to cleanup directory {directory}", UserWarning)

    def __del__(self):
        self.ZMQ_CONTEXT.term()


global_config = Configuration()
"""
Allows to auto apply common settings used throughout the package, instead of passing these settings as arguments.
Import `global_config` variable instead of instantiating this class.

```python
from hololinked.config import global_config

global_config.TEMP_DIR = "/my/temp/dir"
global_config.ALLOW_CORS = True
global_config.setup()  # Important to call setup() after changing values

class MyThing(Thing):
    ...
```

Values are not type checked and are usually mutable in runtime, except:

- logging setup
- ZMQ context
- global event loop policy

which refresh the global state. Keys of JSON file must correspond to supported value name.

Supported values are -

`TEMP_DIR` - system temporary directory to store temporary files like IPC sockets.
default - `~/.hololinked` (`.hololinked` under home directory).

`TCP_SOCKET_SEARCH_START_PORT` - starting port number for automatic port searching
for TCP socket binding, used for event addresses. default `60000`.

`TCP_SOCKET_SEARCH_END_PORT` - ending port number for automatic port searching
for TCP socket binding, used for event addresses. default `65535`.

`DB_CONFIG_FILE` - file path for database configuration. default `None`.

`USE_UVLOOP` - signicantly faster event loop for Linux systems. Reads data from network faster. default `False`.

`TRACE_MALLOC` - whether to trace memory allocations using tracemalloc module. default `False`.

`VALIDATE_SCHEMAS` - whether to validate JSON schema supplied for properties, actions and events
(not validation of payload, but validation of schema itself). default `True`.

`DEBUG` - whether to print debug logs. default `False`.

`LOG_LEVEL` - logging level to use. default `logging.INFO`, `logging.DEBUG` if `DEBUG` is `True`.

`COLORED_LOGS` - whether to use colored logs in console. default `False`.

`ALLOW_PICKLE` - whether to allow pickle serialization/deserialization. default `False`.

`ALLOW_UNKNOWN_SERIALIZATION` - whether to allow unknown serialization formats, specifically from clients. default `False`.
"""

__all__ = ["global_config", "Configuration"]



import pytest


try:
    from .test_11_rpc_e2e import TestRPC_E2E as BaseRPC_E2E  # noqa: F401
    from .test_11_rpc_e2e import client, thing, thing_model  # noqa: F401
except ImportError:
    from test_11_rpc_e2e import TestRPC_E2E as BaseRPC_E2E  # noqa: F401
    from test_11_rpc_e2e import client, thing, thing_model  # noqa: F401


@pytest.fixture(scope="class")
def access_point(request):
    return "IPC"


@pytest.mark.asyncio(loop_scope="class")
class TestZMQ_IPC_E2E(BaseRPC_E2E):
    pass


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



import ssl
import threading
import warnings

from typing import Any

import aiomqtt
import httpx
import structlog

from paho.mqtt.client import CallbackAPIVersion, MQTTMessage, MQTTProtocolVersion
from paho.mqtt.client import Client as PahoMQTTClient

from ..constants import ZMQ_TRANSPORTS
from ..core import Thing
from ..core.zmq import AsyncZMQClient, SyncZMQClient
from ..serializers import Serializers
from ..td.interaction_affordance import (
    ActionAffordance,
    EventAffordance,
    PropertyAffordance,
)
from ..utils import uuid_hex
from .abstractions import ConsumedThingAction, ConsumedThingEvent, ConsumedThingProperty
from .http.consumed_interactions import HTTPAction, HTTPEvent, HTTPProperty
from .mqtt.consumed_interactions import MQTTConsumer  # only one type for now
from .proxy import ObjectProxy
from .security import BasicSecurity, OAuth2Security, OAuthDirectAccessGrant
from .zmq.consumed_interactions import (
    ReadMultipleProperties,
    WriteMultipleProperties,
    ZMQAction,
    ZMQEvent,
    ZMQProperty,
)


class ClientFactory:
    """
    A factory class for creating clients to interact with `Thing`s over different protocols.
    This object is not meant to be instantiated, but rather provides class methods for creating clients.

    ```python
    zmq_client = ClientFactory.zmq(server_id="server1", thing_id="thing1", access_point="ipc:///tmp/thing1")
    http_client = ClientFactory.http(url="https://example.com/thing-description")
    mqtt_client = ClientFactory.mqtt(hostname="broker.example.com", port=8883, thing_id="thing1", username="user", password="pass")
    ```
    """

    __wrapper_assignments__ = ("__name__", "__qualname__", "__doc__")

    @classmethod
    def zmq(
        self,
        server_id: str,
        thing_id: str,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        **kwargs,
    ) -> ObjectProxy:
        """
        Create a ZMQ client for the specified server and thing.

        Parameters
        ----------
        server_id: str
            The ID of the server to connect to
        thing_id: str
            The ID of the thing to interact with
        access_point: str
            The ZMQ protocol to use for communication (`IPC` or `INPROC`) or `tcp://<host>:<port>` for TCP
        kwargs:
            Additional configuration options:

            - `logger`: `structlog.stdlib.BoundLogger`, optional.
                 A custom logger instance to use for logging
            - `ignore_TD_errors`: `bool`, default `False`.
                Whether to ignore errors while fetching the Thing Description (TD)
            - `skip_interaction_affordances`: `list[str]`, default `[]`.
                A list of interaction names to skip (property, action or event names)
            - `invokation_timeout`: `float`, optional, default `5.0`.
                The timeout for invokation requests (in seconds)
            - `execution_timeout`: `float`, optional, default `5.0`.
                The timeout for execution requests (in seconds)

        Returns
        -------
        ObjectProxy
            An ObjectProxy instance representing the remote Thing
        """
        id = kwargs.get("id", f"{server_id}|{thing_id}|{access_point}|{uuid_hex()}")

        # configs
        ignore_TD_errors = kwargs.get("ignore_TD_errors", False)
        skip_interaction_affordances = kwargs.get("skip_interaction_affordances", [])
        invokation_timeout = kwargs.get("invokation_timeout", 5.0)
        execution_timeout = kwargs.get("execution_timeout", 5.0)
        logger = kwargs.get("logger", structlog.get_logger()).bind(
            component="client",
            client_id=id,
            protocol="zmq",
            thing_id=thing_id,
        )

        # ZMQ req-rep clients
        sync_zmq_client = SyncZMQClient(f"{id}|sync", server_id=server_id, logger=logger, access_point=access_point)
        async_zmq_client = AsyncZMQClient(f"{id}|async", server_id=server_id, logger=logger, access_point=access_point)

        # Fetch the TD
        Thing.get_thing_model  # type: Action
        FetchTDAffordance = Thing.get_thing_model.to_affordance()
        FetchTDAffordance.override_defaults(name="get_thing_description", thing_id=thing_id)
        FetchTD = ZMQAction(
            resource=FetchTDAffordance,
            sync_client=sync_zmq_client,
            async_client=async_zmq_client,
            invokation_timeout=invokation_timeout,
            execution_timeout=execution_timeout,
            owner_inst=None,
            logger=logger,
        )
        TD = FetchTD(
            ignore_errors=ignore_TD_errors,
            protocol=access_point.split("://")[0].upper() if access_point else "IPC",
            skip_names=skip_interaction_affordances,
        )  # dict[str, Any]

        # create ObjectProxy
        object_proxy = ObjectProxy(
            id=id,
            td=TD,
            logger=logger,
            invokation_timeout=invokation_timeout,
            execution_timeout=execution_timeout,
            security=kwargs.get("security", None),
        )

        # add properties
        for name in TD.get("properties", []):
            affordance = PropertyAffordance.from_TD(name, TD)
            consumed_property = ZMQProperty(
                resource=affordance,
                sync_client=sync_zmq_client,
                async_client=async_zmq_client,
                owner_inst=object_proxy,
                invokation_timeout=invokation_timeout,
                execution_timeout=execution_timeout,
                logger=logger,
            )
            self.add_property(object_proxy, consumed_property)
            if hasattr(affordance, "observable") and affordance.observable:
                consumed_observable = ZMQEvent(
                    resource=affordance,
                    owner_inst=object_proxy,
                    logger=logger,
                )
                self.add_event(object_proxy, consumed_observable)
        # add actions
        for action in TD.get("actions", []):
            affordance = ActionAffordance.from_TD(action, TD)
            consumed_action = ZMQAction(
                resource=affordance,
                sync_client=sync_zmq_client,
                async_client=async_zmq_client,
                owner_inst=object_proxy,
                invokation_timeout=invokation_timeout,
                execution_timeout=execution_timeout,
                logger=logger,
            )
            self.add_action(object_proxy, consumed_action)
        # add events
        for event in TD.get("events", []):
            affordance = EventAffordance.from_TD(event, TD)
            consumed_event = ZMQEvent(
                resource=affordance,
                owner_inst=object_proxy,
                logger=logger,
            )
            self.add_event(object_proxy, consumed_event)
        # add top level form handlers (for ZMQ even if said form exists or not)
        for opname, ophandler in zip(
            ["_get_properties", "_set_properties"],
            [ReadMultipleProperties, WriteMultipleProperties],
        ):
            setattr(
                object_proxy,
                opname,
                ophandler(
                    sync_client=sync_zmq_client,
                    async_client=async_zmq_client,
                    owner_inst=object_proxy,
                    invokation_timeout=invokation_timeout,
                    execution_timeout=execution_timeout,
                    logger=logger,
                ),
            )
        return object_proxy

    @classmethod
    def http(self, url: str, **kwargs) -> ObjectProxy:
        """
        Create a HTTP client using the Thing Description (TD) available at the specified URL.

        Parameters
        ----------
        url: str
            The URL of the Thing Description (TD) to fetch.
        kwargs:
            Additional configuration options:

            - `logger`: `structlog.stdlib.BoundLogger`, optional.
                A custom logger instance to use for logging
            - `ignore_TD_errors`: `bool`, default `False`.
                Whether to ignore errors while fetching the Thing Description (TD)
            - `skip_interaction_affordances`: `list[str]`, default `[]`.
                A list of interaction names to skip (property, action or event names)
            - `invokation_timeout`: `float`, optional, default `5.0`.
                The timeout for operation invokation (in seconds)
            - `execution_timeout`: `float`, optional, default `5.0`.
                The timeout for operation execution (in seconds)
            - `connect_timeout`: `float`, optional, default `10.0`.
                The timeout for establishing a HTTP connection (in seconds)
            - `request_timeout`: `float`, optional, default `60.0`.
                The timeout for completing a HTTP request (in seconds)
            - `security`: `BasicSecurity` | `APIKeySecurity`, optional.
                The security scheme to use for authentication
            - `username`: `str`, optional.
                The username for HTTP Basic Authentication, shortcut for creating a `BasicSecurity` instance
            - `password`: `str`, optional.
                The password for HTTP Basic Authentication, shortcut for creating a `BasicSecurity` instance

        Returns
        -------
        ObjectProxy
            An ObjectProxy instance representing the remote Thing
        """

        # config
        skip_interaction_affordances = kwargs.get("skip_interaction_affordances", [])
        invokation_timeout = kwargs.get("invokation_timeout", 5.0)
        execution_timeout = kwargs.get("execution_timeout", 5.0)
        connect_timeout = kwargs.get("connect_timeout", 10.0)
        request_timeout = kwargs.get("request_timeout", 60.0)
        use_localhost = False
        if (
            "http://localhost" in url
            or "http://localhost" in url
            or "http://[::1]" in url
            or "http://[::1]" in url
            or "http://127.0.0.1" in url
        ):
            use_localhost = True

        # create clients
        req_rep_timeout = httpx.Timeout(
            connect=connect_timeout,
            read=request_timeout,
            write=request_timeout,
            pool=2,
        )
        sse_timeout = httpx.Timeout(
            connect=connect_timeout,
            read=3,
            write=request_timeout,
            pool=2,
        )

        req_rep_sync_client = httpx.Client(timeout=req_rep_timeout)
        req_rep_async_client = httpx.AsyncClient(timeout=req_rep_timeout)
        sse_sync_client = httpx.Client(timeout=sse_timeout)
        sse_async_client = httpx.AsyncClient(timeout=sse_timeout)

        # fetch TD
        url = (
            f"{url}?"
            + f"ignore_errors={str(kwargs.get('ignore_TD_errors', False)).lower()}"
            + (f"&skip_names={','.join(skip_interaction_affordances)}" if skip_interaction_affordances else "")
            + f"&use_localhost={str(use_localhost).lower()}"
        )

        # fetch TD
        headers = {"Content-Type": "application/json"}
        security = kwargs.pop("security", None)
        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)
        if not security and username and password:
            security = BasicSecurity(username=username, password=password)
        if security and not isinstance(security, OAuthDirectAccessGrant):
            headers[security.http_header_name] = security.http_header
        if security and isinstance(security, OAuthDirectAccessGrant):
            security = OAuth2Security(
                security,
                req_rep_sync_client=req_rep_sync_client,
                req_rep_async_client=req_rep_async_client,
            )
            security.login()
            headers[security.http_header_name] = security.http_header

        response = req_rep_sync_client.get(url, headers=headers)  # type: httpx.Response
        response.raise_for_status()

        TD = Serializers.json.loads(response.content)
        id = kwargs.get("id", f"client|{TD['id']}|HTTP|{uuid_hex()}")
        logger = kwargs.get("logger", structlog.get_logger()).bind(
            component="client",
            client_id=id,
            protocol="http",
            thing_id=TD["id"],
        )
        object_proxy = ObjectProxy(id, td=TD, logger=logger, security=security, **kwargs)

        for name in TD.get("properties", []):
            affordance = PropertyAffordance.from_TD(name, TD)
            consumed_property = HTTPProperty(
                resource=affordance,
                sync_client=req_rep_sync_client,
                async_client=req_rep_async_client,
                invokation_timeout=invokation_timeout,
                execution_timeout=execution_timeout,
                owner_inst=object_proxy,
                logger=logger,
            )
            self.add_property(object_proxy, consumed_property)
            if affordance.observable:
                consumed_event = HTTPEvent(
                    resource=affordance,
                    sync_client=sse_sync_client,
                    async_client=sse_async_client,
                    invokation_timeout=invokation_timeout,
                    execution_timeout=execution_timeout,
                    owner_inst=object_proxy,
                    logger=logger,
                )
                self.add_event(object_proxy, consumed_event)
        for action in TD.get("actions", []):
            affordance = ActionAffordance.from_TD(action, TD)
            consumed_action = HTTPAction(
                resource=affordance,
                sync_client=req_rep_sync_client,
                async_client=req_rep_async_client,
                invokation_timeout=invokation_timeout,
                execution_timeout=execution_timeout,
                owner_inst=object_proxy,
                logger=logger,
            )
            self.add_action(object_proxy, consumed_action)
        for event in TD.get("events", []):
            affordance = EventAffordance.from_TD(event, TD)
            consumed_event = HTTPEvent(
                resource=affordance,
                sync_client=sse_sync_client,
                async_client=sse_async_client,
                invokation_timeout=invokation_timeout,
                execution_timeout=execution_timeout,
                owner_inst=object_proxy,
                logger=logger,
            )
            self.add_event(object_proxy, consumed_event)

        return object_proxy

    @classmethod
    def mqtt(
        self,
        hostname: str,
        port: int,
        thing_id: str,
        protocol_version: MQTTProtocolVersion = MQTTProtocolVersion.MQTTv5,
        qos: int = 1,
        username: str = None,
        password: str = None,
        ssl_context: ssl.SSLContext = None,
        **kwargs,
    ) -> ObjectProxy:
        """
        Create an MQTT client for the specified broker.

        Parameters
        ----------
        hostname: str
            The hostname of the MQTT broker
        port: int
            The port of the MQTT broker
        thing_id: str
            The ID of the thing to interact with
        protocol_version: paho.mqtt.client.MQTTProtocolVersion
            The MQTT protocol version (e.g., MQTTv5)
        qos: int
            The Quality of Service level for MQTT messages (0, 1, or 2)
        username: str, optional
            The username for authenticating with MQTT broker
        password: str, optional
            The password for authenticating with MQTT broker
        kwargs:
            Additional configuration options:

            - `logger`: `structlog.stdlib.BoundLogger`, optional.
                 A custom logger instance to use for logging
        """
        id = kwargs.get("id", f"mqtt-client|{hostname}:{port}|{uuid_hex()}")
        logger = kwargs.get("logger", structlog.get_logger()).bind(
            component="client",
            client_id=id,
            protocol="mqtt",
            thing_id=thing_id,
        )

        td_received_event = threading.Event()
        TD = None

        def fetch_td(client: PahoMQTTClient, userdata, message: MQTTMessage) -> None:
            nonlocal TD, thing_id, logger
            if message.topic != f"{thing_id}/thing-description":
                return
            TD = Serializers.json.loads(message.payload)
            td_received_event.set()

        def on_connect(
            client: PahoMQTTClient,
            userdata: Any,
            flags: Any,
            reason_code: list,
            properties: dict[str, Any],
        ) -> None:  # TODO fix signature
            nonlocal qos
            client.subscribe(f"{thing_id}/#", qos=qos)

        sync_client = PahoMQTTClient(
            callback_api_version=CallbackAPIVersion.VERSION2,
            client_id=id,
            clean_session=True if not protocol_version == MQTTProtocolVersion.MQTTv5 else None,
            protocol=protocol_version,
        )
        if username and password:
            sync_client.username_pw_set(username=username, password=password)
        if ssl_context is not None:
            sync_client.tls_set_context(ssl_context)
        elif kwargs.get("ca_certs", None):
            sync_client.tls_set(ca_certs=kwargs.get("ca_certs", None))
        sync_client.on_connect = on_connect
        sync_client.on_message = fetch_td
        sync_client.connect(hostname, port)
        sync_client.loop_start()

        td_received_event.wait(timeout=10)
        if not TD:
            raise TimeoutError("Timeout while fetching Thing Description (TD) over MQTT")

        if not sync_client._ssl_context and port != 1883:
            warnings.warn(
                "MQTT used without TLS, if you intended to use TLS with a recognised CA & you saw this warning, considering "
                + "opening an issue at https://github.com/hololinked-dev/hololinked. ",
                category=RuntimeWarning,
            )

        async_client = aiomqtt.Client(
            hostname=hostname,
            port=port,
            username=username,
            password=password,
            protocol=protocol_version,
            tls_context=sync_client._ssl_context,
        )

        object_proxy = ObjectProxy(id=id, logger=logger, td=TD)

        for name in TD.get("properties", []):
            affordance = PropertyAffordance.from_TD(name, TD)
            consumed_property = MQTTConsumer(
                sync_client=sync_client,
                async_client=async_client,
                resource=affordance,
                qos=qos,
                logger=logger,
                owner_inst=object_proxy,
            )
            self.add_property(object_proxy, consumed_property)
        for name in TD.get("events", []):
            affordance = EventAffordance.from_TD(name, TD)
            consumed_event = MQTTConsumer(
                sync_client=sync_client,
                async_client=async_client,
                resource=affordance,
                qos=qos,
                logger=logger,
                owner_inst=object_proxy,
            )
            self.add_event(object_proxy, consumed_event)

        return object_proxy

    @classmethod
    def add_action(self, client, action: ConsumedThingAction) -> None:
        """add action to client instance"""
        setattr(action, "__name__", action.resource.name)
        setattr(action, "__qualname__", f"{client.__class__.__name__}.{action.resource.name}")
        setattr(
            action,
            "__doc__",
            action.resource.description or "Invokes the action {} on the remote Thing".format(action.resource.name),
        )
        setattr(client, action.resource.name, action)

    @classmethod
    def add_property(self, client, property: ConsumedThingProperty) -> None:
        """add property to client instance"""
        setattr(property, "__name__", property.resource.name)
        setattr(property, "__qualname__", f"{client.__class__.__name__}.{property.resource.name}")
        setattr(
            property,
            "__doc__",
            property.resource.description
            or "Represents the property {} on the remote Thing".format(property.resource.name),
        )
        setattr(client, property.resource.name, property)

    @classmethod
    def add_event(cls, client, event: ConsumedThingEvent) -> None:
        """add event to client instance"""
        setattr(event, "__name__", event.resource.name)
        setattr(event, "__qualname__", f"{client.__class__.__name__}.{event.resource.name}")
        setattr(
            event,
            "__doc__",
            event.resource.description or "Represents the event {} on the remote Thing".format(event.resource.name),
        )
        if hasattr(event.resource, "observable") and event.resource.observable:
            setattr(client, f"{event.resource.name}_change_event", event)
        else:
            setattr(client, event.resource.name, event)



import logging

import pytest

from pydantic import BaseModel

from hololinked.constants import ResourceTypes
from hololinked.core.properties import (
    Boolean,
    ClassSelector,
    List,
    Number,
    Property,
    Selector,
    String,
)
from hololinked.td.data_schema import DataSchema
from hololinked.td.interaction_affordance import (
    ActionAffordance,
    EventAffordance,
    InteractionAffordance,
    PropertyAffordance,
)
from hololinked.utils import issubklass, uuid_hex


try:
    from .things import OceanOpticsSpectrometer, TestThing
    from .things.spectrometer import Intensity
except ImportError:
    from things import OceanOpticsSpectrometer, TestThing
    from things.spectrometer import Intensity


@pytest.fixture(scope="module")
def thing():
    return OceanOpticsSpectrometer(id=f"test-thing-{uuid_hex()}", log_level=logging.ERROR)


@pytest.fixture(scope="module")
def test_thing():
    return TestThing(id=f"test-spectrometer-thing-{uuid_hex()}", log_level=logging.ERROR)


def test_01_associated_objects(thing):
    affordance = PropertyAffordance()
    affordance.objekt = OceanOpticsSpectrometer.integration_time
    affordance.owner = thing
    assert isinstance(affordance, BaseModel)
    assert isinstance(affordance, DataSchema)
    assert isinstance(affordance, InteractionAffordance)
    assert affordance.what == ResourceTypes.PROPERTY
    assert affordance.owner == thing
    assert affordance.thing_id == thing.id
    assert affordance.thing_cls == thing.__class__
    assert isinstance(affordance.objekt, Property)
    assert affordance.name == OceanOpticsSpectrometer.integration_time.name

    affordance = PropertyAffordance()
    assert affordance.owner is None
    assert affordance.objekt is None
    assert affordance.name is None
    assert affordance.thing_id is None
    assert affordance.thing_cls is None

    affordance = ActionAffordance()
    with pytest.raises(ValueError) as ex:
        affordance.objekt = OceanOpticsSpectrometer.integration_time
    with pytest.raises(TypeError) as ex:
        affordance.objekt = 5
    assert "objekt must be instance of Property, Action or Event, given type" in str(ex.value)
    affordance.objekt = OceanOpticsSpectrometer.connect
    assert affordance.what == ResourceTypes.ACTION

    affordance = EventAffordance()
    with pytest.raises(ValueError) as ex:
        affordance.objekt = OceanOpticsSpectrometer.integration_time
    with pytest.raises(TypeError) as ex:
        affordance.objekt = 5
    assert "objekt must be instance of Property, Action or Event, given type" in str(ex.value)
    affordance.objekt = OceanOpticsSpectrometer.intensity_measurement_event
    assert affordance.what == ResourceTypes.EVENT

    affordance = PropertyAffordance()
    with pytest.raises(ValueError) as ex:
        affordance.objekt = OceanOpticsSpectrometer.connect
    with pytest.raises(TypeError) as ex:
        affordance.objekt = 5
    assert "objekt must be instance of Property, Action or Event, given type" in str(ex.value)
    affordance.objekt = OceanOpticsSpectrometer.integration_time


def test_02_number_schema(thing):
    schema = OceanOpticsSpectrometer.integration_time.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert schema.type == "number"

    integration_time = Number(
        bounds=(1, 1000),
        default=100,
        crop_to_bounds=True,
        step=1,
        doc="integration time in milliseconds",
        metadata=dict(unit="ms"),
    )
    integration_time.__set_name__(OceanOpticsSpectrometer, "integration_time")
    schema = integration_time.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert schema.type == "number"
    assert schema.minimum == integration_time.bounds[0]
    assert schema.maximum == integration_time.bounds[1]
    assert schema.multipleOf == integration_time.step
    with pytest.raises(AttributeError):
        _ = schema.exclusiveMinimum
    with pytest.raises(AttributeError):
        _ = schema.exclusiveMaximum
    integration_time.inclusive_bounds = (False, False)
    integration_time.step = None
    schema = integration_time.to_affordance(owner_inst=thing)
    assert schema.exclusiveMinimum == integration_time.bounds[0]
    assert schema.exclusiveMaximum == integration_time.bounds[1]
    with pytest.raises(AttributeError):
        _ = schema.minimum
    with pytest.raises(AttributeError):
        _ = schema.maximum
    with pytest.raises(AttributeError):
        _ = schema.multipleOf
    integration_time.allow_None = True
    schema = integration_time.to_affordance(owner_inst=thing)
    assert any(subtype["type"] == "null" for subtype in schema.oneOf)
    assert any(subtype["type"] == "number" for subtype in schema.oneOf)
    assert len(schema.oneOf) == 2
    assert not hasattr(schema, "type") or schema.type is None
    number_schema = next(subtype for subtype in schema.oneOf if subtype["type"] == "number")
    assert number_schema["exclusiveMinimum"] == integration_time.bounds[0]
    assert number_schema["exclusiveMaximum"] == integration_time.bounds[1]
    with pytest.raises(KeyError):
        _ = number_schema["minimum"]
    with pytest.raises(KeyError):
        _ = number_schema["maximum"]
    with pytest.raises(KeyError):
        _ = number_schema["multipleOf"]
    assert schema.default == integration_time.default
    assert schema.unit == integration_time.metadata["unit"]


def test_03_string_schema(thing):
    schema = OceanOpticsSpectrometer.status.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)

    status = String(
        regex=r"^[a-zA-Z0-9]{1,10}$",
        default="IDLE",
        doc="status of the spectrometer",
    )
    status.__set_name__(OceanOpticsSpectrometer, "status")
    schema = status.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert schema.type == "string"
    assert schema.pattern == status.regex
    status.allow_None = True
    schema = status.to_affordance(owner_inst=thing)
    assert any(subtype["type"] == "null" for subtype in schema.oneOf)
    assert any(subtype["type"] == "string" for subtype in schema.oneOf)
    assert len(schema.oneOf) == 2
    assert not hasattr(schema, "type") or schema.type is None
    string_schema = next(subtype for subtype in schema.oneOf if subtype["type"] == "string")
    assert string_schema["pattern"] == status.regex
    assert schema.default == status.default


def test_04_boolean_schema(thing):
    schema = OceanOpticsSpectrometer.nonlinearity_correction.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)

    nonlinearity_correction = Boolean(default=True, doc="nonlinearity correction enabled")
    nonlinearity_correction.__set_name__(OceanOpticsSpectrometer, "nonlinearity_correction")
    schema = nonlinearity_correction.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert schema.type == "boolean"
    nonlinearity_correction.allow_None = True
    schema = nonlinearity_correction.to_affordance(owner_inst=thing)
    assert any(subtype["type"] == "null" for subtype in schema.oneOf)
    assert any(subtype["type"] == "boolean" for subtype in schema.oneOf)
    assert len(schema.oneOf) == 2
    assert not hasattr(schema, "type") or schema.type is None
    assert schema.default == nonlinearity_correction.default


def test_05_array_schema(thing):
    schema = OceanOpticsSpectrometer.wavelengths.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)

    wavelengths = List(
        default=[],
        item_type=(float, int),
        readonly=True,
        allow_None=False,
        doc="wavelength bins of measurement",
    )
    wavelengths.__set_name__(OceanOpticsSpectrometer, "wavelengths")
    schema = wavelengths.to_affordance(owner_inst=thing)
    assert isinstance(schema, BaseModel)
    assert isinstance(schema, DataSchema)
    assert isinstance(schema, PropertyAffordance)
    assert schema.type == "array"
    for types in schema.items["oneOf"]:
        assert types["type"] == "number" or types["type"] == "integer"
    if OceanOpticsSpectrometer.wavelengths.default is not None:
        assert schema.default == OceanOpticsSpectrometer.wavelengths.default
    OceanOpticsSpectrometer.wavelengths.allow_None = True
    schema = OceanOpticsSpectrometer.wavelengths.to_affordance(owner_inst=thing)
    assert any(subtype["type"] == "null" for subtype in schema.oneOf)
    assert any(subtype["type"] == "array" for subtype in schema.oneOf)
    assert len(schema.oneOf) == 2
    assert not hasattr(schema, "type") or schema.type is None
    array_schema = next(subtype for subtype in schema.oneOf if subtype["type"] == "array")
    for types in array_schema["items"]["oneOf"]:
        assert types["type"] == "number" or types["type"] == "integer"

    for bounds in [(5, 1000), (None, 100), (50, None), (51, 101)]:
        wavelengths.bounds = bounds
        wavelengths.allow_None = False
        schema = wavelengths.to_affordance(owner_inst=thing)
        if bounds[0] is not None:
            assert schema.minItems == bounds[0]
        else:
            assert not hasattr(schema, "minItems") or schema.minItems is None
        if bounds[1] is not None:
            assert schema.maxItems == bounds[1]
        else:
            assert not hasattr(schema, "maxItems") or schema.maxItems is None
        wavelengths.bounds = bounds
        wavelengths.allow_None = True
        schema = wavelengths.to_affordance(owner_inst=thing)
        subtype = next(subtype for subtype in schema.oneOf if subtype["type"] == "array")
        if bounds[0] is not None:
            assert subtype["minItems"] == bounds[0]
        else:
            with pytest.raises(KeyError):
                _ = subtype["minItems"]
        if bounds[1] is not None:
            assert subtype["maxItems"] == bounds[1]
        else:
            with pytest.raises(KeyError):
                _ = subtype["maxItems"]


def test_06_enum_schema(thing):
    schema = OceanOpticsSpectrometer.trigger_mode.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)

    trigger_mode = Selector(
        objects=[0, 1, 2, 3, 4],
        default=0,
        observable=True,
        doc="""0 = normal/free running, 1 = Software trigger, 2 = Ext. Trigger Level,
                    3 = Ext. Trigger Synchro/ Shutter mode, 4 = Ext. Trigger Edge""",
    )
    trigger_mode.__set_name__(OceanOpticsSpectrometer, "trigger_mode")
    schema = trigger_mode.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert schema.type == "integer"
    assert schema.default == 0
    assert schema.enum == trigger_mode.objects

    trigger_mode.allow_None = True
    trigger_mode.default = 3
    trigger_mode.objects = [0, 1, 2, 3, 4, "0", "1", "2", "3", "4"]
    schema = trigger_mode.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert not hasattr(schema, "type") or schema.type is None
    assert schema.default == 3
    enum_subschema = next(
        subtype
        for subtype in schema.oneOf
        if (subtype.get("type", None) != "null" or len(subtype.get("oneOf", [])) > 1)
    )
    assert isinstance(enum_subschema, dict)
    assert enum_subschema["enum"] == trigger_mode.objects


def test_07_class_selector_custom_schema(thing):
    last_intensity = ClassSelector(
        default=Intensity([], []),
        allow_None=False,
        class_=Intensity,
        doc="last measurement intensity (in arbitrary units)",
    )
    last_intensity.__set_name__(OceanOpticsSpectrometer, "last_intensity")
    schema = last_intensity.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert schema.type == "object"
    assert schema.properties == Intensity.schema["properties"]

    last_intensity.allow_None = True
    schema = last_intensity.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert not hasattr(schema, "type") or schema.type is None
    subschema = next(subtype for subtype in schema.oneOf if subtype.get("type", None) == "object")
    assert isinstance(subschema, dict)
    assert subschema["type"] == "object"
    assert subschema["properties"] == Intensity.schema["properties"]


def test_08_json_schema_properties(thing):
    json_schema_prop = TestThing.json_schema_prop  # type: Property
    json_schema_prop.allow_None = False
    schema = json_schema_prop.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    for key in json_schema_prop.model:
        assert getattr(schema, key, NotImplemented) == json_schema_prop.model[key]

    json_schema_prop.allow_None = True
    schema = json_schema_prop.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    subschema = next(
        subtype
        for subtype in schema.oneOf
        if (subtype.get("type", None) != "null" or len(subtype.get("oneOf", [])) > 1)
    )
    assert isinstance(subschema, dict)
    for key in json_schema_prop.model:
        assert subschema.get(key, NotImplemented) == json_schema_prop.model[key]


def test_09_pydantic_properties(thing):
    pydantic_prop = TestThing.pydantic_prop  # type: Property
    pydantic_prop.allow_None = False
    schema = pydantic_prop.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    if issubklass(pydantic_prop.model, BaseModel):
        assert schema.type == "object"
        for field in pydantic_prop.model.model_fields:
            assert field in schema.properties

    pydantic_prop.allow_None = True
    schema = pydantic_prop.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    subschema = next(subtype for subtype in schema.oneOf if subtype.get("type", None) == "object")
    assert isinstance(subschema, dict)
    for key in pydantic_prop.model.model_fields:
        assert key in subschema.get("properties", {})

    pydantic_simple_prop = TestThing.pydantic_simple_prop  # type: Property # its an integer
    pydantic_simple_prop.allow_None = False
    schema = pydantic_simple_prop.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    assert schema.type == "integer"

    pydantic_simple_prop.allow_None = True
    schema = pydantic_simple_prop.to_affordance(owner_inst=thing)
    assert isinstance(schema, PropertyAffordance)
    subschema = next(subtype for subtype in schema.oneOf if subtype.get("type", None) == "integer")
    assert subschema["type"] == "integer"
    subschema = next(subtype for subtype in schema.oneOf if subtype.get("type", None) == "null")
    assert subschema["type"] == "null"


def test_10_thing_model_generation():
    thing = TestThing(id="test-thing-model", log_level=logging.ERROR + 10)
    assert isinstance(thing.get_thing_model(skip_names=["base_property"]).json(), dict)


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



from typing import Any

from pydantic import BaseModel, Field

from ..repository import BrokerThing, thing_repository  # noqa: F401
from ..security import Security
from .controllers import (
    ActionHandler,
    EventHandler,
    LivenessProbeHandler,
    PropertyHandler,
    ReadinessProbeHandler,
    RPCHandler,
    RWMultiplePropertiesHandler,
    StopHandler,
    ThingDescriptionHandler,
)
from .services import ThingDescriptionService


class RuntimeConfig(BaseModel):
    """
    Runtime configuration for HTTP server and handlers.
    Pass the attributes of this class as a dictionary to the `config` argument of `HTTPServer`.
    """

    cors: bool = False
    """use `True` to set CORS headers for the HTTP server, this is useful for local networks"""

    property_handler: type[RPCHandler] | Any = PropertyHandler
    """handler class to be used for property interactions"""
    action_handler: type[RPCHandler] | Any = ActionHandler
    """handler class to be used for action interactions"""
    event_handler: type[EventHandler] | Any = EventHandler
    """handler class to be used for event interactions"""
    RW_multiple_properties_handler: type[RPCHandler] | Any = RWMultiplePropertiesHandler
    """handler class to be used for read/write multiple properties interactions"""
    thing_description_handler: type[ThingDescriptionHandler] | Any = ThingDescriptionHandler
    """handler class to be used for thing description"""
    liveness_probe_handler: type[LivenessProbeHandler] | Any = LivenessProbeHandler
    """handler class to be used for liveness probe"""
    readiness_probe_handler: type[ReadinessProbeHandler] | Any = ReadinessProbeHandler
    """handler class to be used for readiness probe"""
    stop_handler: type[StopHandler] | Any = StopHandler
    """handler class to be used for stopping server"""

    thing_description_service: type[ThingDescriptionService] | Any = ThingDescriptionService
    """service class to be used for generating thing description"""

    thing_repository: Any = thing_repository  # type: dict[str, BrokerThing]
    """repository layer thing model to be used by the HTTP server and handlers"""

    allowed_clients: list[str] | None = Field(default=None)
    """
    Serves request and sets CORS only from these clients, other clients are rejected with 401. 
    Unlike pure CORS, the server resource is not even executed if the client is not 
    an allowed client. if None, any client is served. Not inherently a safety feature in public networks, 
    and more useful in private networks when the remote origin is known reliably.
    """

    security_schemes: list[Security] | None = Field(default=None)
    """
    List of security schemes to be used by the server, 
    it is sufficient that one scheme passes for a request to be authorized.
    Combo security schemes are not yet supported (but will be in future).
    """


class HandlerMetadata(BaseModel):
    """Specific metadata when a request handler has been initialized, in other words, handler specific metadata"""

    http_methods: tuple[str, ...] = tuple()
    """HTTP methods supported by the handler"""



"""
Classes used to support string serialization of Parameters and
Parameterized objects.
"""

import json
import textwrap


class UnserializableException(Exception):
    pass


class UnsafeserializableException(Exception):
    pass


def JSONNullable(json_type):
    "Express a JSON schema type as nullable to easily support Parameters that allow_None"
    return {"anyOf": [json_type, {"type": "null"}]}


class Serialization(object):
    """
    Base class used to implement different types of serialization.
    """

    @classmethod
    def schema(cls, pobj, subset=None):
        raise NotImplementedError  # noqa: unimplemented method

    @classmethod
    def serialize_parameters(cls, pobj, subset=None):
        """
        Serialize the parameters on a Parameterized object into a
        single serialized object, e.g. a JSON string.
        """
        raise NotImplementedError  # noqa: unimplemented method

    @classmethod
    def deserialize_parameters(cls, pobj, serialized, subset=None):
        """
        Deserialize a serialized object representing one or
        more Parameters into a dictionary of parameter values.
        """
        raise NotImplementedError  # noqa: unimplemented method

    @classmethod
    def serialize_parameter_value(cls, pobj, pname):
        """
        Serialize a single parameter value.
        """
        raise NotImplementedError  # noqa: unimplemented method

    @classmethod
    def deserialize_parameter_value(cls, pobj, pname, value):
        """
        Deserialize a single parameter value.
        """
        raise NotImplementedError  # noqa: unimplemented method


class JSONSerialization(Serialization):
    """
    Class responsible for specifying JSON serialization, deserialization
    and JSON schemas for Parameters and Parameterized classes and
    objects.
    """

    unserializable_parameter_types = ["Callable"]

    json_schema_literal_types = {int: "integer", float: "number", str: "string", type(None): "null"}

    @classmethod
    def loads(cls, serialized):
        return json.loads(serialized)

    @classmethod
    def dumps(cls, obj):
        return json.dumps(obj)

    @classmethod
    def schema(cls, pobj, safe=False, subset=None):
        schema = {}
        for name, p in pobj.param.objects("existing").items():
            if subset is not None and name not in subset:
                continue
            schema[name] = p.schema(safe=safe)
            if p.doc:
                schema[name]["description"] = textwrap.dedent(p.doc).replace("\n", " ").strip()
            if p.label:
                schema[name]["title"] = p.label
        return schema

    @classmethod
    def serialize_parameters(cls, pobj, subset=None):
        # components = {}
        # for name, p in pobj.param.objects('existing').items():
        #     if subset is not None and name not in subset:
        #         continue
        #     value = pobj.param.get_value_generator(name)
        #     components[name] = p.serialize(value)
        # return cls.dumps(components)
        JSON = {}
        pobjtype = type(pobj)
        for key, param in pobj.parameters.objects().items():
            if subset is not None and key not in subset:
                pass
            else:
                value = param.__get__(pobj, pobjtype)
                value = param.serialize(value)
                JSON[key] = value
        return JSON

    @classmethod
    def deserialize_parameters(cls, pobj, serialization, subset=None):
        deserialized = cls.loads(serialization)
        components = {}
        for name, value in deserialized.items():
            if subset is not None and name not in subset:
                continue
            deserialized = pobj.param[name].deserialize(value)
            components[name] = deserialized
        return components

    # Parameter level methods

    @classmethod
    def _get_method(cls, ptype, suffix):
        "Returns specialized method if available, otherwise None"
        method_name = ptype.lower() + "_" + suffix
        return getattr(cls, method_name, None)

    @classmethod
    def param_schema(cls, ptype, p, safe=False, subset=None):
        if ptype in cls.unserializable_parameter_types:
            raise UnserializableException
        dispatch_method = cls._get_method(ptype, "schema")
        if dispatch_method:
            schema = dispatch_method(p, safe=safe)
        else:
            schema = {"type": ptype.lower()}
        return JSONNullable(schema) if p.allow_None else schema

    @classmethod
    def serialize_parameter_value(cls, pobj, pname):
        value = pobj.param.get_value_generator(pname)
        return cls.dumps(pobj.param[pname].serialize(value))

    @classmethod
    def deserialize_parameter_value(cls, pobj, pname, value):
        value = cls.loads(value)
        return pobj.param[pname].deserialize(value)

    # Custom Schemas

    @classmethod
    def class__schema(cls, class_, safe=False):
        from .parameterized import Parameterized

        if isinstance(class_, tuple):
            return {"anyOf": [cls.class__schema(cls_) for cls_ in class_]}
        elif class_ in cls.json_schema_literal_types:
            return {"type": cls.json_schema_literal_types[class_]}
        elif issubclass(class_, Parameterized):
            return {"type": "object", "properties": class_.param.schema(safe)}
        else:
            return {"type": "object"}

    @classmethod
    def array_schema(cls, p, safe=False):
        if safe is True:
            msg = "Array is not guaranteed to be safe for serialization as the dtype is unknown"
            raise UnsafeserializableException(msg)
        return {"type": "array"}

    @classmethod
    def classselector_schema(cls, p, safe=False):
        return cls.class__schema(p.class_, safe=safe)

    @classmethod
    def dict_schema(cls, p, safe=False):
        if safe is True:
            msg = "Dict is not guaranteed to be safe for serialization as the key and value types are unknown"
            raise UnsafeserializableException(msg)
        return {"type": "object"}

    @classmethod
    def date_schema(cls, p, safe=False):
        return {"type": "string", "format": "date-time"}

    @classmethod
    def calendardate_schema(cls, p, safe=False):
        return {"type": "string", "format": "date"}

    @classmethod
    def tuple_schema(cls, p, safe=False):
        schema = {"type": "array"}
        if p.length is not None:
            schema["minItems"] = p.length
            schema["maxItems"] = p.length
        return schema

    @classmethod
    def number_schema(cls, p, safe=False):
        schema = {"type": p.__class__.__name__.lower()}
        return cls.declare_numeric_bounds(schema, p.bounds, p.inclusive_bounds)

    @classmethod
    def declare_numeric_bounds(cls, schema, bounds, inclusive_bounds):
        "Given an applicable numeric schema, augment with bounds information"
        if bounds is not None:
            (low, high) = bounds
            if low is not None:
                key = "minimum" if inclusive_bounds[0] else "exclusiveMinimum"
                schema[key] = low
            if high is not None:
                key = "maximum" if inclusive_bounds[1] else "exclusiveMaximum"
                schema[key] = high
        return schema

    @classmethod
    def integer_schema(cls, p, safe=False):
        return cls.number_schema(p)

    @classmethod
    def numerictuple_schema(cls, p, safe=False):
        schema = cls.tuple_schema(p, safe=safe)
        schema["additionalItems"] = {"type": "number"}
        return schema

    @classmethod
    def xycoordinates_schema(cls, p, safe=False):
        return cls.numerictuple_schema(p, safe=safe)

    @classmethod
    def range_schema(cls, p, safe=False):
        schema = cls.tuple_schema(p, safe=safe)
        bounded_number = cls.declare_numeric_bounds({"type": "number"}, p.bounds, p.inclusive_bounds)
        schema["additionalItems"] = bounded_number
        return schema

    @classmethod
    def list_schema(cls, p, safe=False):
        schema = {"type": "array"}
        if safe is True and p.item_type is None:
            msg = "List without a class specified cannot be guaranteed to be safe for serialization"
            raise UnsafeserializableException(msg)
        if p.class_ is not None:
            schema["items"] = cls.class__schema(p.item_type, safe=safe)
        return schema

    @classmethod
    def objectselector_schema(cls, p, safe=False):
        try:
            allowed_types = [{"type": cls.json_schema_literal_types[type(obj)]} for obj in p.objects]
            schema = {"anyOf": allowed_types}
            schema["enum"] = p.objects
            return schema
        except:
            if safe is True:
                msg = (
                    "ObjectSelector cannot be guaranteed to be safe for "
                    "serialization due to unserializable type in objects"
                )
                raise UnsafeserializableException(msg)
            return {}

    @classmethod
    def selector_schema(cls, p, safe=False):
        try:
            allowed_types = [{"type": cls.json_schema_literal_types[type(obj)]} for obj in p.objects.values()]
            schema = {"anyOf": allowed_types}
            schema["enum"] = p.objects
            return schema
        except:
            if safe is True:
                msg = "Selector cannot be guaranteed to be safe for serialization due to unserializable type in objects"
                raise UnsafeserializableException(msg)
            return {}

    @classmethod
    def listselector_schema(cls, p, safe=False):
        if p.objects is None:
            if safe is True:
                msg = "ListSelector cannot be guaranteed to be safe for serialization as allowed objects unspecified"
            return {"type": "array"}
        for obj in p.objects:
            if type(obj) not in cls.json_schema_literal_types:
                msg = "ListSelector cannot serialize type %s" % type(obj)
                raise UnserializableException(msg)
        return {"type": "array", "items": {"enum": p.objects}}

    @classmethod
    def dataframe_schema(cls, p, safe=False):
        schema = {"type": "array"}
        if safe is True:
            msg = "DataFrame is not guaranteed to be safe for serialization as the column dtypes are unknown"
            raise UnsafeserializableException(msg)
        if p.columns is None:
            schema["items"] = {"type": "object"}
            return schema

        mincols, maxcols = None, None
        if isinstance(p.columns, int):
            mincols, maxcols = p.columns, p.columns
        elif isinstance(p.columns, tuple):
            mincols, maxcols = p.columns

        if isinstance(p.columns, int) or isinstance(p.columns, tuple):
            schema["items"] = {"type": "object", "minItems": mincols, "maxItems": maxcols}

        if isinstance(p.columns, list) or isinstance(p.columns, set):
            literal_types = [{"type": el} for el in cls.json_schema_literal_types.values()]
            allowable_types = {"anyOf": literal_types}
            properties = {name: allowable_types for name in p.columns}
            schema["items"] = {"type": "object", "properties": properties}

        minrows, maxrows = None, None
        if isinstance(p.rows, int):
            minrows, maxrows = p.rows, p.rows
        elif isinstance(p.rows, tuple):
            minrows, maxrows = p.rows

        if minrows is not None:
            schema["minItems"] = minrows
        if maxrows is not None:
            schema["maxItems"] = maxrows

        return schema


serializers = dict(json=JSONSerialization)



import inspect
import logging
import ssl

from typing import Any

import structlog

from ..constants import ZMQ_TRANSPORTS
from ..serializers import Serializers
from ..utils import forkable, getattr_without_descriptor_read
from .actions import BoundAction, action
from .events import EventDispatcher
from .meta import EventSource, Propertized, RemoteInvokable, ThingMeta
from .properties import ClassSelector, String
from .property import Property


class Thing(Propertized, RemoteInvokable, EventSource, metaclass=ThingMeta):
    """
    Subclass from here to expose hardware or python objects on the network. Remotely accessible members of a `Thing` are
    segregated into properties, actions & events. Utilize properties for data that can be read and written,
    actions to instruct the object to perform tasks and events to get notified of any relevant information. State Machines
    can be used to constrain operations on properties and actions.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/Thing.pdf)
    """

    # local properties
    id = String(
        default=None,
        regex=r"[A-Za-z]+[A-Za-z_0-9\-\/]*",
        constant=True,
        remote=False,
        doc="""String identifier of the instance. For an interconnected system of hardware, 
            IDs are recommended to be unique. This value is used for many operations,
            for example - creating zmq socket address, tables in databases, and to identify the instance 
            in the HTTP Server - (http(s)://{domain and sub domain}/{id}).""",
    )  # type: str
    """
    String identifier of the instance. For an interconnected system of hardware, 
    IDs are recommended to be unique. This value is used for many operations,
    for example - creating zmq socket address, tables in databases, and to identify the instance 
    in the HTTP Server - (http(s)://{domain}/{id}).
    """
    # TODO use docstring in only one place

    logger = ClassSelector(
        class_=(logging.Logger, structlog.stdlib.BoundLoggerBase),
        default=None,
        allow_None=True,
        remote=False,
        doc="""structlog.stdlib.BoundLogger instance to log messages. Default logger with a IO-stream handler 
            is created if none supplied.""",
    )  # type: structlog.stdlib.BoundLogger
    """
    structlog.stdlib.BoundLogger instance to log messages. Default logger with a IO-stream handler 
    is created if none supplied.
    """

    state_machine = None  # type: "StateMachine" | None
    """class property reserved for state machine instance if any."""

    # remote properties
    state = String(
        default=None,
        allow_None=True,
        readonly=True,
        observable=True,
        fget=lambda self: self.state_machine.current_state if self.state_machine else None,
        doc="""Current state machine's state if state machine present, `None` indicates absence of state machine.
            State machine returned state is always a string even if specified as an Enum in the state machine.""",
    )  # type: str | None
    """
    Current state machine's state if state machine present, `None` indicates absence of state machine.
    State machine returned state is always a string even if specified as an Enum in the state machine.
    """

    # object_info = Property(doc="contains information about this object like the class name, script location etc.") # type: ThingInformation

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        # defines some internal fixed attributes. attributes created by us that require no validation but
        # cannot be modified are called _internal_fixed_attributes
        obj._internal_fixed_attributes = ["_internal_fixed_attributes", "_owners"]
        return obj

    def __init__(
        self,
        *,
        id: str,
        logger: structlog.stdlib.BoundLogger | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        """
        Parameters
        ----------
        id: str
            String identifier of the instance. For an interconnected system of hardware,
            IDs are recommended to be unique. This value is used for many operations, for example -
            creating zmq socket address, tables in databases, and to identify the instance in a
            HTTP Server - (http(s)://{domain and sub domain}/{id}).
        logger: structlog.stdlib.BoundLogger, optional
            `structlog.stdlib.BoundLogger` instance to log messages. Default logger with a IO-stream handler
            is created if None supplied.
        **kwargs: dict[str, Any]

            - `serializer`: `BaseSerializer` | `JSONSerializer`, optional
                Default serializer to be used for serializing and deserializing data.
                If not supplied, a `msgspec` based JSON Serializer is used.
            - `remote_accessible_logger`: `bool`, Default `False`.
                if `True`, the log records can be streamed by a remote client. `remote_accessible_logger` can also be set as a
                class attribute. Use this a minimalistic replacement for fluentd or similar log collectors.
            - `use_default_db`: `bool`, Default `False`.
                if `True`, default SQLite database is created where properties can be stored and loaded. There is no need to supply
                any database credentials. `use_default_db` value can also be set as a class attribute.
            - `db_config_file`: `str`, optional.
                if not using a default database, supply a JSON configuration file to create a database connection. Check documentaion
                of [`hololinked.core.database`](https://docs.hololinked.dev/api-reference/namespaces/).
            - `use_json_file`: `bool`, Default `False`
                if `True`, a JSON file will be used as the property storage instead of a database. This value can also be
                set as a class attribute.
            - `json_filename`: `str`, optional
                If using JSON storage, this filename is used to persist property values. If not provided, a default filename
                is generated based on the instance name.
        """
        Propertized.__init__(self, id=id, logger=logger, **kwargs)
        RemoteInvokable.__init__(self)
        EventSource.__init__(self)
        if self.id.startswith("/"):
            self.id = self.id[1:]
        if kwargs.get("serializer", None) is not None:
            Serializers.register_for_thing_instance(self.id, kwargs.get("serializer"))

        from .logger import prepare_object_logger  # noqa
        from .state_machine import prepare_object_FSM  # noqa
        from ..storage import prepare_object_storage  # noqa

        prepare_object_logger(
            instance=self,
            remote_access=kwargs.get(
                "remote_accessible_logger",
                self.__class__.remote_accessible_logger
                if hasattr(self.__class__, "remote_accessible_logger")
                else False,
            ),
        )
        prepare_object_FSM(self)
        prepare_object_storage(self, **kwargs)  # use_default_db, db_config_file, use_json_file, json_filename

        self._qualified_id = self.id  # filler for now - TODO
        # thing._qualified_id = f'{self._qualified_id}/{thing.id}'

    def __post_init__(self):
        from ..storage.database import ThingDB
        from .logger import RemoteAccessHandler
        from .zmq.rpc_server import RPCServer  # noqa: F401

        # Type definitions
        self.rpc_server = None  # type: RPCServer | None
        self.db_engine: ThingDB | None
        self._owners = None if not hasattr(self, "_owners") else self._owners  # type: list[Thing] | None
        self._remote_access_loghandler: RemoteAccessHandler | None
        self._internal_fixed_attributes: list[str]
        self._qualified_id: str
        self._state_machine_state: str
        # database operations
        self.properties.load_from_DB()
        # object is ready
        self.logger.info(f"initialialised Thing class {self.__class__.__name__} with id {self.id}")

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "_internal_fixed_attributes" or __name in self._internal_fixed_attributes:
            # order of 'or' operation for above 'if' matters
            if not hasattr(self, __name) or getattr(self, __name, None) is None:
                # allow setting of fixed attributes once
                super().__setattr__(__name, __value)
            else:
                raise AttributeError(
                    f"Attempted to set {__name} more than once. "
                    + "Cannot assign a value to this variable after creation."
                )
        else:
            super().__setattr__(__name, __value)

    @property
    def sub_things(self) -> dict[str, "Thing"]:
        """other `Thing`s' that are composed within this `Thing`."""
        things = dict()
        for name, subthing in inspect._getmembers(
            self,
            lambda obj: isinstance(obj, Thing),
            getattr_without_descriptor_read,  # noqa: F405
        ):
            if not hasattr(subthing, "_owners") or subthing._owners is None:
                subthing._owners = []
            if self not in subthing._owners:
                subthing._owners.append(self)
            things[name] = subthing
        return things

    @action()
    def get_thing_model(self, ignore_errors: bool = False, skip_names: list[str] = []):
        """
        generate the [Thing Model](https://www.w3.org/TR/wot-thing-description11/#introduction-tm) of the object.
        The model is a JSON that describes the object's properties, actions, events and their metadata, without the
        protocol information. The model can be used by a client to understand the object's capabilities.

        Parameters
        ----------
        ignore_errors: `bool`, optional, Default `False`
            if `True`, offending interaction affordances will be removed from the JSON
            (i.e. those who have wrong metadata or non-JSON metadata).
            This is useful to build partial but always working `ThingModel`.
        skip_names: `list[str]`, optional
            List of affordances names (of any type) to skip in the generated model.

        Returns
        -------
        hololinked.td.ThingModel
            represented as an object in python, gets automatically serialized to JSON when pushed out of the socket.
        """
        # allow_loose_schema: bool, optional, Default False
        #     Experimental properties, actions or events for which schema was not given will be supplied with a suitable
        #     inaccurate but truthy value. In other words, schema validation will always pass.
        from ..td.tm import ThingModel

        return ThingModel(instance=self, ignore_errors=ignore_errors, skip_names=skip_names).generate()

    thing_model = property(get_thing_model, doc=get_thing_model.__doc__)

    @forkable  # noqa: F405
    def run_with_zmq_server(
        self,
        access_points: list[ZMQ_TRANSPORTS] | ZMQ_TRANSPORTS | str | list[str] = ZMQ_TRANSPORTS.IPC,
        forked: bool = False,  # used by decorator
        **kwargs: dict[str, Any],
    ) -> None:
        """
        Quick-start to serve `Thing` over ZMQ. This method is fully blocking, call `exit()` (`hololinked.server.exit()`)
        to unblock.

        Parameters
        ----------
        access_points: list[ZMQ_TRANSPORTS] | ZMQ_TRANSPORTS | str | list[str], Default ZMQ_TRANSPORTS.IPC or "IPC"
            ZMQ transport layers at which the object is exposed:

            - `TCP` -  custom implemented protocol in plain TCP - supply a socket address additionally in the format
            `tcp://*:<port>` (for example - `tcp://*:5555`) or a random port will be automatically used.
            The star `*` indicates that the server will listen on all available network interfaces.
            - `IPC` - inter process communication - connection can be made from other processes running
            locally within same computer. No client on the network will be able to contact the object using
            this transport. Use this transport if you wish to avoid configuring your firewall for a local client,
            like a Desktop application or a local web server.
            - `INPROC` - one main python process spawns several threads in one of which the `Thing`
            will be running. The object can be contacted by a client on another thread but not from other processes
            or the network.

            One may use more than one form of transport.  All requests made will be anyway queued internally
            irrespective of origin. For multiple transports, supply a list of transports.
            For example: `[ZMQ_TRANSPORTS.TCP, ZMQ_TRANSPORTS.IPC]`,
            `["TCP", "IPC"]`, `["tcp://*:5555", "IPC"]` or `["IPC", "INPROC"]`.

        forked: bool, Default `False`
            if `True`, the server is started in a separate thread and this method returns immediately.

        **kwargs:
            - context: `zmq.asyncio.Context`, optional,
                ZMQ context object to be used for creating sockets. If not supplied, a global shared context is used.
                For INPROC, either do not supply context or use the same context across all threads.
        """
        from ..server.server import parse_params, run

        servers = parse_params(self.id, [("ZMQ", dict(access_points=access_points, logger=self.logger, **kwargs))])

        for server in servers:
            server.add_thing(self)
        run(*servers, print_welcome_message=False)  # no welcome message for ZMQ

    @forkable  # noqa: F405
    def run_with_http_server(
        self,
        port: int = 8080,
        address: str = "0.0.0.0",  # SAST(id='hololinked.core.thing.Thing.run_with_http_server.address', description='B104:hardcoded_bind_all_interfaces', tool='bandit')
        # host: str = None,
        allowed_clients: str | list[str] | None = None,
        ssl_context: ssl.SSLContext | None = None,
        # protocol_version : int = 1,
        # network_interface : str = 'Ethernet',
        forked: bool = False,  # used by forkable decorator
        print_welcome_message: bool = True,
        **kwargs: dict[str, Any],
    ) -> None:
        """
        Quick-start to serve `Thing` over HTTP. This method is fully blocking, call `exit()` (`hololinked.server.exit()`)
        to unblock.

        Parameters
        ----------
        port: int
            the port at which the HTTP server should be run (unique)
        address: str
            A convenience option to set IP address apart from 0.0.0.0 (i.e. bind to all interfaces, which is default)
        ssl_context: ssl.SSLContext | None
            provide custom certificates with an SSL context for encrypted communication
        allowed_clients: str | list[str] | None
            serves request and sets CORS only from these clients, other clients are rejected with 403. Uses remote IP
            header value to achieve this. Unlike CORS, the server resource is not even executed if the client is not an allowed client.
            Note that the remote IP in a HTTP request is believable only from a trusted HTTP client, not a modified one.
        forked: bool, Default `False`
            if `True`, the server is started in a separate thread and this method returns immediately
        **kwargs: dict[str, Any]
            additional keyword arguments:

            - `property_handler`: `BaseHandler` | `PropertyHandler`,
                custom web request handler for property operations
            - `action_handler`: `BaseHandler` | `ActionHandler`,
                custom web request handler for action operations
            - `event_handler`: `BaseHandler` | `EventHandler`,
                custom event handler for handling events
        """
        from ..server import HTTPServer, run

        http_server = HTTPServer(
            port=port,
            address=address,
            logger=self.logger,
            ssl_context=ssl_context,
            allowed_clients=allowed_clients,
            **kwargs,
        )
        http_server.add_thing(self)
        run(http_server, print_welcome_message=print_welcome_message)

    @forkable  # noqa: F405
    def run(
        self,
        forked: bool = False,  # used by forkable decorator
        print_welcome_message: bool = True,
        **kwargs: dict[str, Any],
    ) -> None:
        """
        Expose the object with the given servers. This method is blocking until `exit()` (`hololinked.server.exit()`)
        is called.

        >>> Thing(id='example_id').run(servers=[http_server, zmq_server, mqtt_publisher])

        Parameters
        ----------
        forked: bool, Default `False`
            if `True`, the server is started in a separate thread and this method returns immediately.
        print_welcome_message: bool, Default `True`
            if `True`, a welcome message with server information is printed to the console.
        kwargs: dict[str, Any]
            keyword arguments

            - `access_points`: dict[str, dict | int | str | list[str]], optional
                The protocol name and its port or parameters to expose the object.
                for example - `[('HTTP', 9000), ('ZMQ', 'tcp://*:5555')]`

            - `servers`: list[BaseProtocolServer]
                list of instantiated servers to expose the object.
        """
        from ..server.server import BaseProtocolServer, parse_params, run  # noqa: F401

        access_points = kwargs.get("access_points", None)  # type: dict[str, dict | int | str | list[str]]
        servers = kwargs.get("servers", [])  # type: list[BaseProtocolServer] | None

        if access_points is None and len(servers) == 0:
            raise ValueError("At least one of access_points or servers must be provided.")
        if access_points is not None and len(servers) > 0:
            raise ValueError("Only one of access_points or servers can be provided.")

        if access_points is not None:
            servers = parse_params(self.id, access_points)
        for server in servers:
            server.add_thing(self)
        run(*servers, print_welcome_message=print_welcome_message)

    @action()
    def exit(self) -> None:
        """
        Stop serving the object. This method usually needs to be called remotely.
        The servers are not stopped, just the object run loop is exited.
        """
        if self.rpc_server is None:
            self.logger.debug("exit() called on a object that is not exposed yet.")
            return
        if self._owners:
            raise NotImplementedError(
                "call exit on the top-level object, composed objects cannot exit the loop. "
                + f"This object belongs to {self._owners.__class__.__name__} with ID {self._owners.id}."
            )
        self.rpc_server.stop()

    @action()
    def ping(self) -> None:
        """
        ping to see if it is alive. Successful when action succeeds with no return value and
        no timeout or exception raised on the client side.
        """
        pass

    def __hash__(self) -> int:
        filename = inspect.getfile(self.__class__)
        if filename is not None:
            # i.e. try to make it as unique as possible
            return hash(filename + self.__class__.__name__ + self.id)
        return hash(self.__class__.__name__ + self.id)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Thing):
            return False
        return self.__class__ == other.__class__ and self.id == other.id

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.id})"

    def __contains__(self, item: Property | BoundAction | EventDispatcher) -> bool:
        return item in self.properties or item in self.actions or item in self.events

    def __enter__(self) -> "Thing":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass


from .state_machine import StateMachine  # noqa: F401, E402


__all__ = [Thing.__name__]



from enum import Enum, EnumMeta, StrEnum
from types import FunctionType, MethodType
from typing import Callable

from ..param import edit_constant
from .actions import Action
from .exceptions import StateMachineError
from .meta import ThingMeta
from .properties import Boolean, ClassSelector, TypedDict
from .property import Property
from .thing import Thing


class StateMachine:
    """
    A finite state machine to constrain property and action execution. Each `Thing` class can only have one state machine
    instantiated in a reserved class-level attribute named `state_machine`. Other instantiations are not respected.
    The `state` attribute defined as a `Thing`'s property reflects the current state of the state machine and
    can be subscribed for state change events. When `state_machine` is accessed by a `Thing` instance,
    a `BoundFSM` object is returned.
    """

    initial_state = ClassSelector(
        default=None,
        allow_None=True,
        constant=True,
        class_=(Enum, str),
        doc="initial state of the machine",
    )  # type: Enum | str
    """initial state of the machine"""

    states = ClassSelector(
        default=None,
        allow_None=True,
        constant=True,
        class_=(EnumMeta, tuple, list),
        doc="list/enum of allowed states",
    )  # type: EnumMeta | tuple | list
    """list of allowed states"""

    on_enter = TypedDict(
        default=None,
        allow_None=True,
        key_type=str,
        doc="""callbacks to execute when a certain state is entered; 
            specified as map with state as keys and callbacks as list""",
    )  # type: dict[str, list[Callable]]
    """
    callbacks to execute when a certain state is entered; 
    specified as map with state as keys and callbacks as list
    """

    on_exit = TypedDict(
        default=None,
        allow_None=True,
        key_type=str,
        doc="""callbacks to execute when certain state is exited; 
            specified as map with state as keys and callbacks as list""",
    )  # type: dict[str, list[Callable]]
    """
    callbacks to execute when certain state is exited; 
    specified as map with state as keys and callbacks as list
    """

    machine = TypedDict(
        default=None,
        allow_None=True,
        item_type=(list, tuple),
        key_type=str,  # i.e. its like JSON
        doc="the machine specification with state as key and objects as list",
    )  # type: dict[str, list[Callable | Property]]
    """the machine specification with state as key and objects as list"""

    push_state_change_event = Boolean(
        default=True,
        doc="if `True`, when the state changes, an event is pushed with the new state",
    )  # type: bool
    """if `True`, when the state changes, an event is pushed with the new state"""

    valid = Boolean(
        default=False,
        readonly=True,
        fget=lambda self: self._valid,
        doc="internally computed, `True` if states, initial_states and the machine is valid",
    )  # type: bool
    """internally computed, `True` if states, initial_states and the machine is valid"""

    def __init__(
        self,
        states: EnumMeta | list[str] | tuple[str],
        *,
        initial_state: StrEnum | str,
        push_state_change_event: bool = True,
        on_enter: dict[str, list[Callable] | Callable] = None,
        on_exit: dict[str, list[Callable] | Callable] = None,
        **machine: dict[str, Callable | Property],
    ) -> None:
        """
        Parameters
        ----------
        states: EnumMeta | List[str] | Tuple[str]
            enumeration of states
        initial_state: StrEnum | str
            initial state of machine
        push_state_change_event: bool, default `True`
            when the state changes, an event is pushed to clients with the new state as the payload
        on_enter: Dict[str, Callable | Property]
            callbacks to be invoked when a certain state is entered. It is to be specified
            as a dictionary with the states being the keys and the list of functions or methods as values.
        on_exit: Dict[str, Callable | Property]
            callbacks to be invoked when a certain state is exited.
            It is to be specified as a dictionary with the states being the keys
            and the list of functions or methods as values.
        **machine:
            the state machine specification with state as key and list of methods or properties as values.

            `state name`: List[Callable, Property]
                directly pass the state name as an argument along with the methods/properties
                which are allowed to execute in that state
        """
        self._valid = False
        self.name = None
        self.on_enter = on_enter
        self.on_exit = on_exit
        # None cannot be passed in, but constant is necessary.
        self.states = states
        self.initial_state = initial_state
        self.machine = machine
        self.push_state_change_event = push_state_change_event
        self.logger = None

    def __set_name__(self, owner: ThingMeta, name: str) -> None:
        self.name = name
        self.owner = owner

    def validate(self, owner: Thing) -> None:
        """validate the state machine, whether the properties, actions and states are correctly specified"""

        # cannot merge this with __set_name__ because descriptor objects are not ready at that time.
        # reason - metaclass __init__ is called after __set_name__ of descriptors, therefore the new "proper" desriptor
        # registries are available only after that. Until then only the inherited descriptor registries are available,
        # which do not correctly account the subclass's objects.
        if self.states is None and self.initial_state is None:
            self._valid = False
            return
        elif self.initial_state not in self.states:
            raise AttributeError(f"specified initial state {self.initial_state} not in Enum of states {self.states}.")

        owner_properties = owner.properties.get_descriptors(recreate=True).values()
        owner_methods = owner.actions.get_descriptors(recreate=True).values()

        if isinstance(self.states, list):
            with edit_constant(self.__class__.states):  # type: ignore
                self.states = tuple(self.states)  # freeze the list of states

        # first validate machine
        for state, objects in self.machine.items():
            if state in self:
                for resource in objects:
                    if isinstance(resource, Action):
                        if resource not in owner_methods:
                            raise AttributeError(
                                "Given object {} for state machine does not belong to class {}".format(resource, owner)
                            )
                    elif isinstance(resource, Property):
                        if resource not in owner_properties:
                            raise AttributeError(
                                "Given object {} for state machine does not belong to class {}".format(resource, owner)
                            )
                        continue  # for now
                    else:
                        raise AttributeError(
                            f"Object {resource} was not made remotely accessible,"
                            + " use state machine with properties and actions only."
                        )
                    if resource.execution_info.state is None:
                        resource.execution_info.state = self._get_machine_compliant_state(state)
                    else:
                        resource.execution_info.state = resource._execution_info.state + (
                            self._get_machine_compliant_state(state),
                        )
            else:
                raise StateMachineError(
                    "Given state {} not in allowed states ({})".format(state, self.states.__members__)
                )

        # then the callbacks
        if self.on_enter is None:
            self.on_enter = {}
        for state, objects in self.on_enter.items():
            if isinstance(objects, list):
                self.on_enter[state] = tuple(objects)
            elif not isinstance(objects, (list, tuple)):
                self.on_enter[state] = (objects,)
            for obj in self.on_enter[state]:  # type: ignore
                if not isinstance(obj, (FunctionType, MethodType)):
                    raise TypeError(f"on_enter accept only methods. Given type {type(obj)}.")

        if self.on_exit is None:
            self.on_exit = {}
        for state, objects in self.on_exit.items():
            if isinstance(objects, list):
                self.on_exit[state] = tuple(objects)  # type: ignore
            elif not isinstance(objects, (list, tuple)):
                self.on_exit[state] = (objects,)  # type: ignore
            for obj in self.on_exit[state]:  # type: ignore
                if not isinstance(obj, (FunctionType, MethodType)):
                    raise TypeError(f"on_enter accept only methods. Given type {type(obj)}.")

        self.logger = owner.logger.bind(component="state-machine", thing_id=owner.id)
        self._valid = True

    def __get__(self, instance, owner) -> "BoundFSM":
        if instance is None:
            return self
        return BoundFSM(instance, self)

    def __set__(self, instance, value) -> None:
        raise AttributeError(
            "Cannot set state machine directly. It is a class level attribute and can be defined only once."
        )

    def __contains__(self, state: str | StrEnum):
        if isinstance(self.states, EnumMeta) and state in self.states.__members__:
            return True
        elif isinstance(self.states, tuple) and state in self.states:
            return True
        return False

    def _get_machine_compliant_state(self, state) -> StrEnum | str:
        """
        In case of not using StrEnum or iterable of str,
        this maps the enum of state to the state name.
        """
        if isinstance(state, str):
            return state
        if isinstance(state, Enum):
            return state.name
        raise TypeError(
            f"cannot comply state to a string: {state} which is of type {type(state)}. owner - {self.owner}."
        )

    def contains_object(self, object: Property | Callable) -> bool:
        """
        Check if specified object is found in any of the state machine states.
        Supply unbound method for checking methods, as state machine is specified at class level
        when the methods are unbound.

        Parameters
        ----------
        object: Property | Callable
            The unbound method or property

        Returns
        -------
        bool
            `True` if the object is found in any of the states, `False` otherwise
        """
        for objects in self.machine.values():
            if object in objects:
                return True
        return False


class BoundFSM:
    """
    A FSM bound to a `Thing` instance, returned when accessed as a instance attribute (`self.state_machine`).
    There is no need to instantiate this class directly.
    """

    def __init__(self, owner: Thing, state_machine: StateMachine) -> None:
        self.descriptor = state_machine
        self.push_state_change_event = state_machine.push_state_change_event
        self.owner = owner
        self.logger = state_machine.logger

    def get_state(self) -> str | StrEnum | None:
        """
        return the current state, one can also access it using the property `current state`.

        Returns
        -------
        str
            current state of the state machine
        """
        try:
            return self.owner._state_machine_state
        except AttributeError:
            return self.initial_state

    def set_state(self, value: str | StrEnum | Enum, push_event: bool = True, skip_callbacks: bool = False) -> None:
        """
        set state of state machine. Also triggers state change callbacks if `skip_callbacks=False` and pushes a state
        change event when `push_event=True` (when __init__ argument `push_state_change_event=True`).
        One can also set state using the '=' operator of the `current_state` property,
        in which case `skip_callbacks=False` and `push_event=True` will be used.

        If originally an enumeration for the list of allowed states was supplied,
        then an enumeration member must be used to set the state. If a list of strings were supplied,
        then a string is accepted.

        Raises
        ------
        ValueError:
            if the state is not found in the allowed states
        """
        if value in self.states:
            given_state = self.descriptor._get_machine_compliant_state(value)
            if given_state == self.current_state:
                return
            previous_state = self.current_state
            next_state = self.descriptor._get_machine_compliant_state(value)
            self.owner._state_machine_state = next_state
            self.logger.info(f"state changed from {previous_state} to {next_state}")
            if push_event and self.push_state_change_event and hasattr(self.owner, "event_publisher"):
                self.owner.state  # just acces to trigger the observable event
            if skip_callbacks:
                return
            if previous_state in self.on_exit:
                for func in self.on_exit[previous_state]:
                    func(self.owner)
            if next_state in self.on_enter:
                for func in self.on_enter[next_state]:
                    func(self.owner)
        else:
            raise ValueError("given state '{}' not in set of allowed states : {}.".format(value, self.states))

    current_state = property(get_state, set_state, None, doc="""read and write current state of the state machine""")

    def contains_object(self, object: Property | Callable) -> bool:
        """
        Check if specified object is found in any of the state machine states.
        Supply unbound method for checking methods, as state machine is specified at class level
        when the methods are unbound.
        """
        return self.descriptor.contains_object(object)

    def __hash__(self):
        return hash(
            self.owner.id
            + (str(state) for state in self.states)
            + str(self.initial_state)
            + self.owner.__class__.__name__
        )

    def __str__(self):
        return f"StateMachine(owner={self.owner.__class__.__name__} id={self.owner.id} initial_state={self.initial_state}, states={self.states})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, StateMachine):
            return False
        return (
            self.states == other.states
            and self.initial_state == other.initial_state
            and self.owner.__class__ == other.owner.__class__
            and self.owner.id == other.owner.id
        )

    def __contains__(self, state: str | StrEnum) -> bool:
        return state in self.descriptor

    @property
    def initial_state(self):
        """initial state of the machine"""
        return self.descriptor.initial_state

    @property
    def states(self):
        """list of allowed states"""
        return self.descriptor.states

    @property
    def on_enter(self):
        """callbacks to execute when a certain state is entered"""
        return self.descriptor.on_enter

    @property
    def on_exit(self):
        """callbacks to execute when certain state is exited"""
        return self.descriptor.on_exit

    @property
    def machine(self):
        """the machine specification with state as key and objects as list"""
        return self.descriptor.machine


def prepare_object_FSM(instance: Thing) -> None:
    """validate and prepare the state machine attached to a Thing class"""
    cls = instance.__class__
    if cls.state_machine and isinstance(cls.state_machine, StateMachine):
        cls.state_machine.validate(instance)
        instance.logger.info(
            f"setup state machine, states={[state.name if hasattr(state, 'name') else state for state in cls.state_machine.states]}, "
            + f"initial_state={cls.state_machine.initial_state}"
        )



import base64
import datetime
import itertools
import json
import os
import random
import sys
import time

from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Generator

import httpx
import pytest
import requests

from pydantic import BaseModel

from hololinked.client import ClientFactory, ObjectProxy
from hololinked.client.security import APIKeySecurity as ClientAPIKeySecurity
from hololinked.client.security import OAuthDirectAccessGrant
from hololinked.config import global_config
from hololinked.core.zmq.message import (
    PreserializedData,
    SerializableData,
    ServerExecutionContext,
    ThingExecutionContext,
    default_server_execution_context,
)
from hololinked.serializers import (
    BaseSerializer,
    JSONSerializer,
    MsgpackSerializer,
    PickleSerializer,
)
from hololinked.server import stop
from hololinked.server.http import HTTPServer, RPCHandler
from hololinked.server.security import (
    APIKeySecurity,
    Argon2BasicSecurity,
    BcryptBasicSecurity,
    OIDCSecurity,
    Security,
)
from hololinked.utils import uuid_hex


try:
    from .things import OceanOpticsSpectrometer
except ImportError:
    from things import OceanOpticsSpectrometer


hostname_prefix = "http://127.0.0.1"
readiness_endpoint = "/readiness"
liveness_endpoint = "/liveness"
stop_endpoint = "/stop"
start_acquisition_endpoint = "/start-acquisition"
intensity_measurement_event_endpoint = "/intensity-measurement-event"
stop_acquisition_endpoint = "/stop-acquisition"

count = itertools.count(62000)


@pytest.fixture(scope="module")
def session() -> requests.Session:
    return requests.Session()


@pytest.fixture(scope="function")
def port() -> int:
    global count
    return next(count)


@pytest.fixture(scope="function")
def server(port) -> Generator[HTTPServer, None, None]:
    server = HTTPServer(address="127.0.0.1", port=port)
    server.run(forked=True, print_welcome_message=False)
    wait_until_server_ready(port=port)
    yield server
    stop()


@pytest.fixture(scope="function")
def thing(port: int) -> Generator[OceanOpticsSpectrometer, None, None]:
    thing = OceanOpticsSpectrometer(id=f"test-thing-{uuid_hex()}", serial_number="simulation")
    print()  # TODO, can be removed when tornado logs respect level
    thing.run_with_http_server(
        address="127.0.0.1",
        port=port,
        forked=True,
        print_welcome_message=False,
        config=dict(cors=True),
    )
    wait_until_server_ready(port=port)
    yield thing
    stop()


@contextmanager
def running_thing(
    id_prefix: str,
    port: int = None,
    **http_server_kwargs,
) -> Generator[OceanOpticsSpectrometer, None, None]:
    """same as thing fixture but to use it manually"""
    global count
    port = port or next(count)
    thing = OceanOpticsSpectrometer(id=f"{id_prefix}-{uuid_hex()}", serial_number="simulation")
    print()  # TODO, can be removed when tornado logs respect level
    thing.run_with_http_server(
        address="127.0.0.1",
        port=port,
        forked=True,
        config=dict(cors=True),
        print_welcome_message=False,
        **http_server_kwargs,
    )
    wait_until_server_ready(port=port)
    try:
        yield thing
    finally:
        stop()


@pytest.fixture(scope="function")
def endpoints(thing: OceanOpticsSpectrometer) -> list[tuple[str, str, Any]]:
    return running_thing_endpoints(thing)


@pytest.fixture(scope="function")
def td_endpoint(thing: OceanOpticsSpectrometer, port: int) -> str:
    return f"{hostname_prefix}:{port}/{thing.id}/resources/wot-td"


@pytest.fixture(scope="function")
def object_proxy(td_endpoint: str) -> "ObjectProxy":
    return ClientFactory.http(url=td_endpoint)


def running_thing_endpoints(thing: OceanOpticsSpectrometer) -> list[tuple[str, str, Any]]:
    if thing.__class__ == OceanOpticsSpectrometer:
        return [
            # properties
            ("get", f"/{thing.id}/max-intensity", 16384),
            ("get", f"/{thing.id}/serial-number", "simulation"),
            ("put", f"/{thing.id}/integration-time", 1200),
            ("get", f"/{thing.id}/integration-time", 1200),
            # actions
            ("post", f"/{thing.id}/disconnect", None),
            ("post", f"/{thing.id}/connect", None),
        ]
    raise NotImplementedError(f"endpoints cannot be generated for {thing.__class__}")


def wait_until_server_ready(port: int, tries: int = 10) -> None:
    session = requests.Session()
    for _ in range(tries):
        try:
            response = session.get(f"{hostname_prefix}:{port}{liveness_endpoint}")
            if response.status_code in [200, 201, 202, 204]:
                response = session.get(f"{hostname_prefix}:{port}{readiness_endpoint}")
                if response.status_code in [200, 201, 202, 204]:
                    return
        except Exception:
            pass
        time.sleep(1)
    print(f"Server on port {port} not ready after {tries} tries, you need to retrigger this test job")
    sys.exit(1)


def sse_stream(url: str, chunk_size: int = 2048, **kwargs):
    with requests.get(url, stream=True, **kwargs) as resp:
        resp.raise_for_status()
        buffer = ""  # type: str
        for chunk in resp.iter_content(chunk_size=chunk_size, decode_unicode=True):
            buffer += chunk
            while "\n\n" in buffer:
                raw_event, buffer = buffer.split("\n\n", 1)
                event = {}
                for line in raw_event.splitlines():
                    if not line or line.startswith(":"):
                        continue
                    if ":" in line:
                        field, value = line.split(":", 1)
                        event.setdefault(field, "")
                        event[field] += value.lstrip()
                yield event


async def test_01_init_run_and_stop(port: int):
    server = HTTPServer(address="127.0.0.1", port=port)
    server.run(forked=True, print_welcome_message=False)
    wait_until_server_ready(port=port)
    await server.async_stop()
    stop()
    time.sleep(2)

    # stop remotely
    server.run(forked=True, print_welcome_message=False)
    wait_until_server_ready(port=port)
    time.sleep(2)
    response = requests.post(f"{hostname_prefix}:{port}{stop_endpoint}")
    assert response.status_code in [200, 201, 202, 204]
    time.sleep(2)
    await server.async_stop()
    stop()


def test_02_add_interaction_affordance(server: HTTPServer):
    server.add_property("/max-intensity", OceanOpticsSpectrometer.max_intensity)
    server.add_action("/connect", OceanOpticsSpectrometer.connect)
    server.add_event("/intensity/event", OceanOpticsSpectrometer.intensity_measurement_event)
    assert "/max-intensity" in server.router
    assert "/connect" in server.router
    assert "/intensity/event" in server.router
    # replacing interaction affordances on an existing URL path causes a warning
    with pytest.warns(UserWarning):
        server.add_property("/max-intensity", OceanOpticsSpectrometer.last_intensity)
    with pytest.warns(UserWarning):
        server.add_action("/connect", OceanOpticsSpectrometer.disconnect)
    with pytest.warns(UserWarning):
        server.add_event("/intensity/event", OceanOpticsSpectrometer.intensity_measurement_event)


# tests 03 & 04 removed as they need more work to be done


class TestableRPCHandler(RPCHandler):
    """
    handler that tests RPC handler functionalities, without executing an operation on a Thing
    Needs to be replaced with a mock
    """

    @dataclass
    class LatestRequestInfo:
        server_execution_context: ServerExecutionContext | dict[str, Any]
        thing_execution_context: ThingExecutionContext | dict[str, Any]
        payload: SerializableData
        preserialized_payload: PreserializedData

    latest_request_info: LatestRequestInfo

    def update_latest_request_info(self) -> None:
        server_execution_context, thing_execution_context, _, _ = self.get_execution_parameters()
        payload, preserialized_payload = self.get_request_payload()
        TestableRPCHandler.latest_request_info = TestableRPCHandler.LatestRequestInfo(
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
            payload=payload,
            preserialized_payload=preserialized_payload,
        )

    async def get(self):
        self.update_latest_request_info()
        self.set_status(200)
        self.finish()

    async def put(self):
        self.update_latest_request_info()
        self.set_status(200)
        self.finish()

    async def post(self):
        await self.handle_through_thing("invokeaction")


@pytest.fixture(scope="function")
def test_rpc_handler_thing(port: int) -> Generator[OceanOpticsSpectrometer, None, None]:
    global_config.ALLOW_PICKLE = True
    with running_thing(
        id_prefix="test-rpc-handler",
        port=port,
        property_handler=TestableRPCHandler,
        action_handler=TestableRPCHandler,
    ) as thing:
        yield thing
    global_config.ALLOW_PICKLE = False


@pytest.mark.parametrize(
    "serializer",
    [
        pytest.param(JSONSerializer(), id="json"),
        pytest.param(MsgpackSerializer(), id="msgpack"),
        pytest.param(PickleSerializer(), id="pickle"),
    ],
)
@pytest.mark.parametrize(
    "endpoint",
    [
        pytest.param(("get", "/integration-time", None), id="get without params"),
        pytest.param(("get", "/integration-time?fetchExecutionLogs=true", None), id="get with fetchExecutionLogs"),
        pytest.param(
            ("get", "/integration-time?fetchExecutionLogs=true&oneway=true", None),
            id="get with fetchExecutionLogs and oneway",
        ),
        pytest.param(
            ("get", "/integration-time?oneway=true&invokationTimeout=100", None),
            id="get with oneway and invokationTimeout",
        ),
        pytest.param(
            (
                "get",
                "/integration-time?invokationTimeout=100&executionTimeout=120&fetchExecutionLogs=true",
                None,
            ),
            id="get with all params",
        ),
        pytest.param(("put", "/integration-time", 1200), id="put without params"),
        pytest.param(
            ("put", "/integration-time?fetchExecutionLogs=true", {"a": 1, "b": 2}), id="put with fetchExecutionLogs"
        ),
        pytest.param(
            ("put", "/integration-time?fetchExecutionLogs=true&oneway=true", [1, 2, 3]),
            id="put with fetchExecutionLogs and oneway",
        ),
        pytest.param(
            ("put", "/integration-time?oneway=true&invokationTimeout=100", "abcd"),
            id="put with oneway and invokationTimeout",
        ),
        pytest.param(
            (
                "put",
                "/integration-time?invokationTimeout=100&executionTimeout=120&fetchExecutionLogs=true",
                True,
            ),
            id="put with all params",
        ),
    ],
)
def test_05_handlers(
    session: requests.Session,
    test_rpc_handler_thing: OceanOpticsSpectrometer,
    port: int,
    serializer: BaseSerializer,
    endpoint: tuple[str, str, Any],
):
    """Test request info and payload decoding in RPC handlers along with content type handling"""

    method, path, body = endpoint
    response = session.request(
        method=method,
        url=f"{hostname_prefix}:{port}/{test_rpc_handler_thing.id}{path}",
        data=serializer.dumps(body) if body is not None else None,
        headers={"Content-Type": serializer.content_type},
    )
    assert response.status_code in [200, 201, 202, 204]
    # test ThingExecutionContext
    assert isinstance(TestableRPCHandler.latest_request_info.thing_execution_context, ThingExecutionContext)
    if "fetchExecutionLogs" in path:
        assert TestableRPCHandler.latest_request_info.thing_execution_context.fetchExecutionLogs
    else:
        assert not TestableRPCHandler.latest_request_info.thing_execution_context.fetchExecutionLogs
    # test ServerExecutionContext
    assert isinstance(TestableRPCHandler.latest_request_info.server_execution_context, ServerExecutionContext)
    if "oneway" in path:
        assert TestableRPCHandler.latest_request_info.server_execution_context.oneway
    else:
        assert not TestableRPCHandler.latest_request_info.server_execution_context.oneway
    if "invokationTimeout" in path:
        assert TestableRPCHandler.latest_request_info.server_execution_context.invokationTimeout == 100
    else:
        assert (
            TestableRPCHandler.latest_request_info.server_execution_context.invokationTimeout
            == default_server_execution_context.invokationTimeout
        )
    if "executionTimeout" in path:
        assert TestableRPCHandler.latest_request_info.server_execution_context.executionTimeout == 120
    else:
        assert (
            TestableRPCHandler.latest_request_info.server_execution_context.executionTimeout
            == default_server_execution_context.executionTimeout
        )
    assert TestableRPCHandler.latest_request_info.payload.deserialize() == body


def do_a_path_e2e(session: requests.Session, endpoint: tuple[str, str, Any], **request_kwargs):
    """
    basic end-to-end test with the HTTP server using handlers.
    Auth & other features not included, only invokation of interaction affordances.
    """
    method, path, body = endpoint
    # request will go through the Thing object
    response = session.request(
        method=method,
        url=path,
        data=JSONSerializer().dumps(body) if body is not None and method != "get" else None,
        **request_kwargs,
    )
    assert response.status_code in [200, 201, 202, 204]
    # check if the response body is as expected
    if body and method != "put":
        assert response.json() == body
    # check headers
    assert "Access-Control-Allow-Origin" in response.headers
    # assert "Access-Control-Allow-Credentials" in response.headers
    assert "Content-Type" in response.headers

    # test unsupported HTTP methods
    response = session.request(
        method="post" if method in ["get", "put"] else random.choice(["put", "delete"]) if method == "post" else method,
        # get and put become post and post becomes put
        # i.e swap the default HTTP method with an unsupported one to generate 405
        url=path,
        data=JSONSerializer().dumps(body) if body is not None and method != "get" else None,
        **request_kwargs,
    )
    assert response.status_code == 405

    # check options for supported HTTP methods
    response = session.options(path, **request_kwargs)
    assert response.status_code in [200, 201, 202, 204]
    assert "Access-Control-Allow-Origin" in response.headers
    # assert "Access-Control-Allow-Credentials" in response.headers
    assert "Access-Control-Allow-Headers" in response.headers
    assert "Access-Control-Allow-Methods" in response.headers
    allow_methods = response.headers.get("Access-Control-Allow-Methods", [])
    assert (  # noqa
        method.upper() in allow_methods,
        f"Method {method} not allowed in {allow_methods}",
    )


def do_a_path_invalid_auth_e2e(session: requests.Session, endpoint: tuple[str, str, Any], headers: dict = None):
    method, path, body = endpoint
    response = session.request(
        method=method,
        url=path,
        data=JSONSerializer().dumps(body) if body is not None and method != "get" else None,
        headers=headers,
    )
    assert response.status_code == 401


def do_authenticated_path_e2e(
    session: requests.Session,
    endpoint: tuple[str, str, Any],
    auth_headers: dict[str, str] = None,
    wrong_auth_headers: list[dict[str, str]] = None,
):
    """Test end-to-end with authentication"""
    do_a_path_e2e(session, endpoint, headers=auth_headers)
    for wrong_auth_header in wrong_auth_headers:
        do_a_path_invalid_auth_e2e(session, endpoint, headers=wrong_auth_header)


def test_06_basic_end_to_end(
    thing: OceanOpticsSpectrometer,
    session: requests.Session,
    port: int,
    endpoints: list[tuple[str, str, Any]],
) -> None:
    """basic end-to-end test with the HTTP server using handlers."""
    for method, path, body in endpoints:
        do_a_path_e2e(
            session=session,
            endpoint=(method, f"{hostname_prefix}:{port}{path}", body),
            headers={"Content-Type": "application/json"},
        )


@pytest.mark.parametrize(
    "security_scheme",
    [
        BcryptBasicSecurity(username="someuser", password="somepassword"),
        Argon2BasicSecurity(username="someuser", password="somepassword"),
    ],
)
def test_07_basic_security_end_to_end(session: requests.Session, port: int, security_scheme: Security):
    """Test end-to-end with Basic Authentication."""
    with running_thing(id_prefix="test-sec", port=port, security_schemes=[security_scheme]) as thing:
        endpoints = running_thing_endpoints(thing)
        for method, path, body in endpoints:
            do_authenticated_path_e2e(
                session=session,
                endpoint=(f"{method}", f"{hostname_prefix}:{port}{path}", body),
                auth_headers={
                    "Content-type": "application/json",
                    "Authorization": f"Basic {base64.b64encode(b'someuser:somepassword').decode('utf-8')}",
                },
                wrong_auth_headers=[
                    {
                        "Content-type": "application/json",
                        "Authorization": f"Basic {base64.b64encode(b'wronguser:wrongpassword').decode('utf-8')}",
                    },
                    {
                        "Content-type": "application/json",
                        "Authorization": f"Basic {base64.b64encode(b'someuser:wrongpassword').decode('utf-8')}",
                    },
                    {
                        "Content-type": "application/json",
                        "Authorization": f"Basic {base64.b64encode(b'wronguser:somepassword').decode('utf-8')}",
                    },
                    {
                        "Content-type": "application/json",
                        # no header
                    },
                ],
            )


def test_08_apikey_security_end_to_end(session: requests.Session, port: int):
    """Test end-to-end with API Key Authentication."""
    keyname = f"e2e-test-apikey-{uuid_hex()}"
    apikey = APIKeySecurity(name=keyname).create(print_value=False)

    security_scheme = APIKeySecurity(name=keyname)
    with running_thing(id_prefix="test-sec", port=port, security_schemes=[security_scheme]) as thing:
        endpoints = running_thing_endpoints(thing)
        for method, path, body in endpoints:
            do_authenticated_path_e2e(
                session=session,
                endpoint=(f"{method}", f"{hostname_prefix}:{port}{path}", body),
                auth_headers={
                    "Content-type": "application/json",
                    "x-api-key": apikey,
                },
                wrong_auth_headers=[
                    {
                        "Content-type": "application/json",
                        "x-api-key": "wrongapikey",
                    },  # plainly wrong key
                    {
                        "Content-type": "application/json",
                        "x-api-key": f"wotdat-{apikey.split('-')[1].split('.')[0]}.blablabla",
                    },  # wrong key with right ID
                    {
                        "Content-type": "application/json",
                        # no header
                    },
                ],
            )
        for method, path, body in endpoints:  # test key expiration
            security_scheme.record.expiry_at = datetime.datetime.fromisoformat("2000-01-01T00:00:00")
            do_a_path_invalid_auth_e2e(
                session,
                (f"{method}", f"{hostname_prefix}:{port}{path}", body),
                headers={
                    "Content-type": "application/json",
                    "x-api-key": apikey,
                },
            )


@pytest.mark.parametrize(
    "security_scheme, headers",
    [
        pytest.param(None, {}, id="no-auth"),
        pytest.param(
            BcryptBasicSecurity(username="someuser", password="somepassword"),
            {
                "Content-type": "application/json",
                "Authorization": f"Basic {base64.b64encode(b'someuser:somepassword').decode('utf-8')}",
            },
            id="bcrypt-basic-auth",
        ),
        pytest.param(
            APIKeySecurity(name="sse-apikey"),
            {
                "Content-type": "application/json",
                "x-api-key": APIKeySecurity(name="sse-apikey").create(print_value=False, override=True),
            },
            id="apikey-auth",
        ),
    ],
)
def test_09_sse(
    session: requests.Session,
    port: int,
    security_scheme: Security | None,
    headers: dict[str, str],
) -> None:
    """Test Server-Sent Events (SSE)"""
    if hasattr(security_scheme, "load"):
        security_scheme.load()  # TODO refactor later, we should not do fixture based specific setup here
    with running_thing(
        id_prefix="test-sse",
        port=port,
        security_schemes=[security_scheme] if security_scheme else None,
    ) as thing:
        response = session.post(f"{hostname_prefix}:{port}/{thing.id}/start-acquisition", headers=headers)
        assert response.status_code == 200
        sse_gen = sse_stream(
            f"{hostname_prefix}:{port}/{thing.id}/intensity-measurement-event",
            headers=headers,
        )
        for _ in range(5):
            evt = next(sse_gen)
            assert "exception" not in evt and "data" in evt
        response = session.post(f"{hostname_prefix}:{port}/{thing.id}/stop-acquisition", headers=headers)
        assert response.status_code == 200


def test_10_forms_generation(session: requests.Session, td_endpoint: str) -> None:
    response = session.get(td_endpoint)

    assert response.status_code == 200
    td = response.json()

    assert "properties" in td
    assert "actions" in td
    assert "events" in td
    assert len(td["properties"]) >= 0
    assert len(td["actions"]) >= 0
    assert len(td["events"]) >= 0
    for interaction in list(td["properties"].values()) + list(td["actions"].values()) + list(td["events"].values()):
        assert "forms" in interaction
        assert len(interaction["forms"]) > 0
        for form in interaction["forms"]:
            assert "href" in form
            assert "htv:methodName" in form
            assert "contentType" in form
            assert "op" in form


async def test_11_object_proxy_basic(object_proxy: ObjectProxy) -> None:
    assert isinstance(object_proxy, ObjectProxy)
    assert object_proxy.test_echo("Hello World!") == "Hello World!"
    assert await object_proxy.async_invoke_action("test_echo", "Hello World!") == "Hello World!"
    assert object_proxy.read_property("max_intensity") == 16384
    assert object_proxy.write_property("integration_time", 1200) is None
    assert object_proxy.read_property("integration_time") == 1200


def test_12_object_proxy_with_basic_auth(port: int) -> None:
    security_scheme = BcryptBasicSecurity(username="cliuser", password="clipass")
    with running_thing(
        id_prefix="test-basic-auth-object-proxy",
        port=port,
        security_schemes=[security_scheme],
    ) as thing:
        td_endpoint = f"{hostname_prefix}:{port}/{thing.id}/resources/wot-td"
        object_proxy = ClientFactory.http(
            url=td_endpoint,
            username="cliuser",
            password="clipass",
        )
        assert len(object_proxy.td["security"]) > 0
        assert security_scheme.name in object_proxy.td["security"]
        assert security_scheme.name in object_proxy.td["securityDefinitions"]

        assert object_proxy.read_property("max_intensity") == 16384

        with pytest.raises(httpx.HTTPStatusError) as excinfo:
            ClientFactory.http(url=td_endpoint)
        assert excinfo.value.response.status_code == 401


def test_13_object_proxy_with_apikey(port: int) -> None:
    keyname = f"cli-test-apikey-{uuid_hex()}"
    apikey = APIKeySecurity(name=keyname).create(print_value=False)
    security_scheme = APIKeySecurity(name=keyname)
    with running_thing(
        id_prefix="test-apikey-object-proxy",
        port=port,
        security_schemes=[security_scheme],
    ) as thing:
        td_endpoint = f"{hostname_prefix}:{port}/{thing.id}/resources/wot-td"
        object_proxy = ClientFactory.http(
            url=td_endpoint,
            security=ClientAPIKeySecurity(value=apikey),
        )
        assert len(object_proxy.td["security"]) > 0
        assert security_scheme.name in object_proxy.td["security"]
        assert security_scheme.name in object_proxy.td["securityDefinitions"]

        assert object_proxy.read_property("max_intensity") == 16384

        with pytest.raises(httpx.HTTPStatusError) as excinfo:
            ClientFactory.http(url=td_endpoint)
        assert excinfo.value.response.status_code == 401


class OIDCConfig(BaseModel):
    issuer: str
    audience: str
    client_id: str
    client_secret: str
    username: str
    password: str


def test_14_object_proxy_with_oidc(port: int) -> None:
    config_b64 = os.getenv("OIDC_TEST_CONFIG_1_B64")
    config = OIDCConfig(**json.loads(base64.b64decode(config_b64).decode("utf-8")))

    # uncomment for local tests and put a config file in said location
    # config = OIDCConfig(**json.loads(open(f"tests{os.sep}helper-scripts{os.sep}oidc-config-2.json").read()))

    oidc_security = OIDCSecurity(
        issuer=config.issuer,
        audience="device-server",
    )
    client_security = OAuthDirectAccessGrant(
        username=config.username,
        password=config.password,
        oidc_config_url=f"{config.issuer}/.well-known/openid-configuration",
        client_id=config.client_id,
        client_secret=config.client_secret,
    )
    with running_thing(
        id_prefix="test-oidc-object-proxy",
        port=port,
        security_schemes=[oidc_security],
    ) as thing:
        td_endpoint = f"{hostname_prefix}:{port}/{thing.id}/resources/wot-td"
        object_proxy = ClientFactory.http(
            url=td_endpoint,
            security=client_security,
        )
        assert len(object_proxy.td["security"]) > 0
        assert oidc_security.name in object_proxy.td["security"]
        assert oidc_security.name in object_proxy.td["securityDefinitions"]

        with pytest.raises(httpx.HTTPStatusError) as excinfo:
            ClientFactory.http(url=td_endpoint)
        assert excinfo.value.response.status_code == 401
        excinfo = None

        assert object_proxy.read_property("max_intensity") == 16384

        object_proxy._security.logout()  # to test logout functionality
        with pytest.raises(httpx.HTTPStatusError) as excinfo:
            object_proxy.read_property("max_intensity")
        assert excinfo.value.response.status_code == 401


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



import copy

from typing import Any

import structlog

from ...constants import Operations
from ...td.interaction_affordance import EventAffordance, PropertyAffordance


class ThingDescriptionService:
    """
    Generates Thing Descriptions for `Thing`s.
    This object would be a service in layered architecture.
    """

    def __init__(
        self,
        hostname: str,
        port: int,
        logger: structlog.stdlib.BoundLogger,
        ssl: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        hostname: str
            The MQTT broker hostname, to fill in the TD forms
        port: int
            The MQTT broker port, to fill in the TD forms
        logger: structlog.stdlib.BoundLogger
            The logger to use for logging messages
        ssl: bool
            Whether the broker is using SSL or not
        """
        self.hostname = hostname
        self.port = port
        self.logger = logger.bind(layer="service", impl=self.__class__.__name__)
        self.ssl = ssl

    async def generate(
        self,
        ZMQ_TD: dict[str, Any],
        ignore_errors: bool = False,
        skip_names: list[str] = [],
    ) -> dict[str, Any]:
        """
        Generates the Thing Description for the specified `Thing`, adds observable properties and events.

        Parameters
        ----------
        ZMQ_TD: dict[str, Any]
            The ZMQ Thing Description message received from ZMQ broker
        ignore_errors: bool
            Whether to ignore errors when adding properties/events to the TD
        skip_names: list[str]
            List of property/event names to skip when adding to the TD

        Returns
        -------
        dict[str, Any]
            The generated MQTT Thing Description
        """
        TD = copy.deepcopy(ZMQ_TD)
        # remove actions as they dont push events
        TD.pop("actions", None)

        self.add_properties(TD, ZMQ_TD, ignore_errors, skip_names)
        self.add_events(TD, ZMQ_TD, ignore_errors, skip_names)

        return TD

    def add_properties(
        self,
        TD: dict[str, Any],
        ZMQ_TD: dict[str, Any],
        ignore_errors: bool = False,
        skip_names: list[str] = [],
    ) -> None:
        """
        Adds observable properties to the Thing Description with MQTT forms.

        Parameters
        ----------
        TD: dict[str, Any]
            The seed Thing Description to modify in place. This method does not have a return value, therefore
            just supply the TD dict and it will be modified. Non-observable properties will be removed.
        ZMQ_TD: dict[str, Any]
            The ZMQ Thing Description message received from ZMQ broker
        ignore_errors: bool
            Whether to ignore errors when adding properties to the TD
        skip_names: list[str]
            List of property names to skip when adding to the TD
        """
        for name in ZMQ_TD.get("properties", {}).keys():
            if name in skip_names:
                continue
            try:
                affordance = PropertyAffordance.from_TD(name, ZMQ_TD)
                if not affordance.observable:
                    TD["properties"].pop(name)
                    continue
                TD["properties"][name]["forms"] = []
                form = affordance.retrieve_form(Operations.observeproperty)
                form.href = f"mqtt{'s' if self.ssl else ''}://{self.hostname}:{self.port}"
                form.mqv_topic = f"{TD['id']}/{name}"
                TD["properties"][name]["forms"].append(form.json())
            except Exception as ex:
                if ignore_errors:
                    self.logger.warning(f"Could not add property {name} to MQTT TD: {ex}")
                    continue
                raise ex from None

    def add_events(
        self,
        TD: dict[str, Any],
        ZMQ_TD: dict[str, Any],
        ignore_errors: bool = False,
        skip_names: list[str] = [],
    ) -> None:
        """
        Adds events to the Thing Description with MQTT forms.

        Parameters
        ----------
        TD: dict[str, Any]
            The seed Thing Description to modify in place. This method does not have a return value, therefore
            just supply the TD dict and it will be modified.
        ZMQ_TD: dict[str, Any]
            The ZMQ Thing Description message received from ZMQ broker
        ignore_errors: bool
            Whether to ignore errors when adding events to the TD
        skip_names: list[str]
            List of event names to skip when adding to the TD
        """
        # repurpose event
        for name in ZMQ_TD.get("events", {}).keys():
            if name in skip_names:
                continue
            try:
                affordance = EventAffordance.from_TD(name, ZMQ_TD)
                TD["events"][name]["forms"] = []
                form = affordance.retrieve_form(Operations.subscribeevent)
                form.href = f"mqtt{'s' if self.ssl else ''}://{self.hostname}:{self.port}"
                form.mqv_topic = f"{TD['id']}/{name}"
                TD["events"][name]["forms"].append(form.json())
            except Exception as ex:
                if ignore_errors:
                    self.logger.warning(f"Could not add event {name} to MQTT TD: {ex}")
                    continue
                raise ex from None



from typing import Any, ClassVar, Optional

from pydantic import BaseModel, ConfigDict, Field, RootModel

from ..constants import JSON, JSONSerializable
from ..core import Property
from ..core.properties import (
    Boolean,
    ClassSelector,
    Filename,
    Foldername,
    Integer,
    List,
    Number,
    Path,
    Selector,
    String,
    Tuple,
    TupleSelector,
    TypedDict,
    TypedKeyMappingsDict,
    TypedList,
)
from ..schema_validators.json_schema import JSONSchema
from ..utils import issubklass
from .base import Schema
from .utils import get_summary


class DataSchema(Schema):
    """
    Implements Data Schema, usually used to represent payloads of properties, actions and events in a
    WoT Thing Description.

    - [Vocabulary Definitions](https://www.w3.org/TR/wot-thing-description11/#sec-data-schema-vocabulary-definition)
    - [Supported Fields](https://www.w3.org/TR/wot-thing-description11/#data-schema-fields)
    """

    title: str = None
    titles: Optional[dict[str, str]] = None
    description: str = None
    descriptions: Optional[dict[str, str]] = None
    const: Optional[bool] = None
    default: Optional[Any] = None
    readOnly: Optional[bool] = None
    writeOnly: Optional[bool] = None
    # write only can be considered as actions with no return value, so not used in this repository
    format: Optional[str] = None
    unit: Optional[str] = None
    type: Optional[str] = None
    oneOf: Optional[list[JSON]] = None

    model_config = ConfigDict(extra="allow")
    _custom_schema_generators: ClassVar = dict()

    def __init__(self):
        super().__init__()

    def ds_build_fields_from_property(self, property: Property) -> None:
        """populates schema information from property descriptor object"""
        self.title = get_summary(property.doc)
        if property.constant:
            self.const = property.constant
        if property.readonly:
            self.readOnly = property.readonly
        if property.default is not None:
            self.default = property.default
        if property.doc:
            self.description = Schema.format_doc(property.doc)
            if self.title and self.description.startswith(self.title):
                self.description.lstrip(self.title)
                self.description.lstrip(".").lstrip()
                self.title = ""
        if property.metadata and property.metadata.get("unit", None) is not None:
            self.unit = property.metadata["unit"]
        if property.allow_None:
            if (hasattr(self, "type") and self.type is not None) or (hasattr(self, "oneOf") and self.oneOf):
                self._move_own_type_to_oneOf()
            if not hasattr(self, "oneOf") or self.oneOf is None:
                self.oneOf = []
            if not any(types.get("type", NotImplemented) in [None, "null"] for types in self.oneOf):
                self.oneOf.append(dict(type="null"))
        if not self.title:
            del self.title

    # & _ds prefix is used to avoid name conflicts with PropertyAffordance class
    # you dont know what you are building, whether the data schema or something else when viewed from property affordance
    def ds_build_from_property(self, property: Property) -> None:
        """
        generates the schema specific to the property type,
        calls `ds_build_fields_from_property()` after choosing the right type
        """
        if self._custom_schema_generators.get(property, NotImplemented) is not NotImplemented:
            data_schema = self._custom_schema_generators[property]()
        elif isinstance(property, Property) and property.model is not None:
            base_data_schema = DataSchema()
            if isinstance(property.model, dict):
                given_data_schema = property.model
            elif issubklass(property.model, (BaseModel, RootModel)):
                from .pydantic_extensions import type_to_dataschema

                given_data_schema = type_to_dataschema(property.model)
            if property.allow_None:
                base_data_schema.oneOf = []
                base_data_schema.oneOf.append(dict(type="null"))
            if base_data_schema.oneOf:  # allow_None = True
                base_data_schema.oneOf.append(given_data_schema)
            else:
                for key, value in given_data_schema.items():
                    setattr(base_data_schema, key, value)
            data_schema = base_data_schema
        elif isinstance(property, (String, Filename, Foldername, Path)):
            data_schema = StringSchema()
        elif isinstance(property, (Number, Integer)):
            data_schema = NumberSchema()
        elif isinstance(property, Boolean):
            data_schema = BooleanSchema()
        elif isinstance(property, (List, TypedList, Tuple, TupleSelector)):
            data_schema = ArraySchema()
        elif isinstance(property, Selector):
            data_schema = EnumSchema()
        elif isinstance(property, (TypedDict, TypedKeyMappingsDict)):
            data_schema = ObjectSchema()
        elif isinstance(property, ClassSelector):
            data_schema = SelectorSchema()
        else:
            raise TypeError(
                f"WoT schema generator for this descriptor/property is not implemented. name {property.name} & type {type(property)}"
            )

        data_schema.ds_build_fields_from_property(property)
        for field_name in data_schema.model_dump(exclude_unset=True).keys():
            field_value = getattr(data_schema, field_name, NotImplemented)
            if field_value is not NotImplemented:
                setattr(self, field_name, field_value)

    def _move_own_type_to_oneOf(self):
        """move a type to oneOf"""
        pass


class BooleanSchema(DataSchema):
    """
    boolean schema - https://www.w3.org/TR/wot-thing-description11/#booleanschema
    used by Boolean descriptor
    """

    def __init__(self):
        super().__init__()

    def ds_build_fields_from_property(self, property) -> None:
        """generates the schema"""
        self.type = "boolean"
        super().ds_build_fields_from_property(property)

    def _move_own_type_to_oneOf(self):
        if not hasattr(self, "type") or self.type is None:
            return
        if not hasattr(self, "oneOf") or self.oneOf is None:
            self.oneOf = []
        self.oneOf.append(dict(type=self.type))
        del self.type


class StringSchema(DataSchema):
    """
    string schema - https://www.w3.org/TR/wot-thing-description11/#stringschema
    used by String, Filename, Foldername, Path descriptors
    """

    pattern: Optional[str] = None
    minLength: Optional[int] = None
    maxLength: Optional[int] = None

    def __init__(self):
        super().__init__()

    def ds_build_fields_from_property(self, property) -> None:
        """generates the schema"""
        self.type = "string"
        if isinstance(property, String):
            if property.regex is not None:
                self.pattern = property.regex
        super().ds_build_fields_from_property(property)

    def _move_own_type_to_oneOf(self):
        if not hasattr(self, "type") or self.type is None:
            return
        if not hasattr(self, "oneOf") or self.oneOf is None:
            self.oneOf = []
        schema = dict(type=self.type)
        del self.type
        if self.pattern is not None:
            schema["pattern"] = self.pattern
            del self.pattern
        self.oneOf.append(schema)


class NumberSchema(DataSchema):
    """
    number schema - https://www.w3.org/TR/wot-thing-description11/#numberschema
    used by Number and Integer descriptors
    """

    minimum: Optional[int | float] = None
    maximum: Optional[int | float] = None
    exclusiveMinimum: Optional[int | float] = None
    exclusiveMaximum: Optional[int | float] = None
    multipleOf: Optional[int | float] = None

    def __init__(self):
        super().__init__()

    def ds_build_fields_from_property(self, property) -> None:
        """generates the schema"""
        if isinstance(property, Integer):
            self.type = "integer"
        elif isinstance(property, Number):  # dont change order - one is subclass of other
            self.type = "number"
        if property.bounds is not None:
            if isinstance(property.bounds[0], (int, float)):  # i.e. value is not None which is allowed by param
                if not property.inclusive_bounds[0]:
                    self.exclusiveMinimum = property.bounds[0]
                else:
                    self.minimum = property.bounds[0]
            if isinstance(property.bounds[1], (int, float)):
                if not property.inclusive_bounds[1]:
                    self.exclusiveMaximum = property.bounds[1]
                else:
                    self.maximum = property.bounds[1]
        if property.step:
            self.multipleOf = property.step
        super().ds_build_fields_from_property(property)

    def _move_own_type_to_oneOf(self):
        if not hasattr(self, "type") or self.type is None:
            return
        if not hasattr(self, "oneOf") or self.oneOf is None:
            self.oneOf = []
        schema = dict(type=self.type)
        del self.type
        for attr in ["minimum", "maximum", "exclusiveMinimum", "exclusiveMaximum", "multipleOf"]:
            if hasattr(self, attr) and getattr(self, attr) is not None:
                schema[attr] = getattr(self, attr)
                delattr(self, attr)
        self.oneOf.append(schema)


class ArraySchema(DataSchema):
    """
    array schema - https://www.w3.org/TR/wot-thing-description11/#arrayschema
    Used by list, Tuple, TypedList and TupleSelector
    """

    items: Optional[DataSchema | list[DataSchema] | JSON | JSONSerializable] = None
    maxItems: Optional[int] = Field(None, ge=0)
    minItems: Optional[int] = Field(None, ge=0)

    def __init__(self):
        super().__init__()

    def ds_build_fields_from_property(self, property) -> None:
        """generates the schema"""
        self.type = "array"
        self.items = []
        if isinstance(property, (List, Tuple, TypedList)) and property.item_type is not None:
            if property.bounds:
                if property.bounds[0]:
                    self.minItems = property.bounds[0]
                if property.bounds[1]:
                    self.maxItems = property.bounds[1]
            if isinstance(property.item_type, (list, tuple)):
                for typ in property.item_type:
                    self.items.append(dict(type=JSONSchema.get_base_type(typ)))
            elif property.item_type is not None:
                self.items.append(dict(type=JSONSchema.get_base_type(property.item_type)))
        elif isinstance(property, TupleSelector):
            objects = list(property.objects)
            for obj in objects:
                if any(types["type"] == JSONSchema._replacements.get(type(obj), None) for types in self.items):
                    continue
                self.items.append(dict(type=JSONSchema.get_base_type(type(obj))))
        if len(self.items) == 0:
            del self.items
        elif len(self.items) > 1:
            self.items = dict(
                oneOf=self.items
            )  # just adds the key 'oneOf' to items if there is more than one item type
        super().ds_build_fields_from_property(property)

    def _move_own_type_to_oneOf(self):
        if not hasattr(self, "type") or self.type is None:
            return
        if not hasattr(self, "oneOf") or self.oneOf is None:
            self.oneOf = []
        schema = dict(type=self.type)
        del self.type
        for attr in ["items", "maxItems", "minItems"]:
            if hasattr(self, attr) and getattr(self, attr) is not None:
                schema[attr] = getattr(self, attr)
                delattr(self, attr)
        self.oneOf.append(schema)


class ObjectSchema(DataSchema):
    """
    object schema - https://www.w3.org/TR/wot-thing-description11/#objectschema
    Used by TypedDict where the key type must be a string
    """

    properties: Optional[JSON] = None
    required: Optional[list[str]] = None

    def __init__(self):
        super().__init__()

    def ds_build_fields_from_property(self, property) -> None:
        """generates the schema"""
        super().ds_build_fields_from_property(property)
        properties = None
        required = None
        if hasattr(property, "json_schema"):
            # Code will not reach here for now as have not implemented schema for typed dictionaries.
            properties = property.json_schema["properties"]
            if property.json_schema.get("required", NotImplemented) is not NotImplemented:
                required = property.json_schema["required"]
        if not property.allow_None:
            self.type = "object"
            if properties:
                self.properties = properties
            if required:
                self.required = required
        else:
            schema = dict(type="object")
            if properties:
                schema["properties"] = properties
            if required:
                schema["required"] = required
            self.oneOf.append(schema)


class SelectorSchema(DataSchema):
    """
    custom schema to deal with ClassSelector & Selector to fill oneOf field correctly
    https://www.w3.org/TR/wot-thing-description11/#dataschema
    """

    properties: Optional[JSON] = None
    required: Optional[list[str]] = None
    items: Optional[JSON | JSONSerializable] = None
    minItems: Optional[int] = None
    maxItems: Optional[int] = None
    # ClassSelector can technically have a JSON serializable as a class_

    def __init__(self):
        super().__init__()

    def ds_build_fields_from_property(self, property) -> None:
        """generates the schema"""
        self.oneOf = []
        if isinstance(property, ClassSelector):
            if not property.isinstance:
                raise NotImplementedError(
                    "WoT TD/TM for ClassSelector with isinstance set to True is not supported yet. "
                    + "Consider using this property in a different way if you want in the TD/TM."
                )
            if isinstance(property.class_, (list, tuple)):
                objects = list(property.class_)
            else:
                objects = [property.class_]
        elif isinstance(property, Selector):
            objects = list(property.objects)
        else:
            raise TypeError(
                f"EnumSchema and SelectorSchema supported only for Selector and ClassSelector. Given Type - {property}"
            )
        for obj in objects:
            if any(types["type"] == JSONSchema._replacements.get(type(obj), None) for types in self.oneOf):
                # just avoid duplicates of same type in oneOf
                continue
            if isinstance(property, ClassSelector):
                if not JSONSchema.is_allowed_type(obj):
                    raise TypeError(
                        f"Object for wot-td has invalid type for JSON conversion. Given type - {obj}. "
                        + "Use JSONSchema.register_replacements on hololinked.wot.td.JSONSchema object to recognise the type."
                    )
                subschema = dict(type=JSONSchema.get_base_type(obj))
                if JSONSchema.has_additional_schema_definitions(obj):
                    subschema.update(JSONSchema.get_additional_schema_definitions(obj))
                self.oneOf.append(subschema)
            elif isinstance(property, Selector):
                if JSONSchema.get_base_type(type(obj)) == "null":
                    continue
                self.oneOf.append(dict(type=JSONSchema.get_base_type(type(obj))))

        super().ds_build_fields_from_property(property)

        if len(self.oneOf) == 1:
            oneOf = self.oneOf[0]
            self.type = oneOf["type"]
            if oneOf["type"] == "object":
                if oneOf.get("properties", NotImplemented) is not NotImplemented:
                    self.properties = oneOf["properties"]
                if oneOf.get("required", NotImplemented) is not NotImplemented:
                    self.required = oneOf["required"]
            elif oneOf["type"] == "array":
                if oneOf.get("items", NotImplemented) is not NotImplemented:
                    self.items = oneOf["items"]
                if oneOf.get("maxItems", NotImplemented) is not NotImplemented:
                    self.minItems = oneOf["minItems"]
                if oneOf.get("maxItems", NotImplemented) is not NotImplemented:
                    self.maxItems = oneOf["maxItems"]
            del self.oneOf

    def _move_own_type_to_oneOf(self):
        pass


class EnumSchema(SelectorSchema):
    """
    custom schema to fill enum field correctly
    https://www.w3.org/TR/wot-thing-description11/#dataschema
    """

    enum: Optional[list[Any]] = None

    def __init__(self):
        super().__init__()

    def ds_build_fields_from_property(self, property) -> None:
        """generates the schema"""
        if not isinstance(property, Selector):
            raise TypeError(f"EnumSchema compatible property is only Selector, not {property.__class__}")
        self.enum = list(property.objects)
        super().ds_build_fields_from_property(property)

    def _move_own_type_to_oneOf(self):
        if hasattr(self, "type") and self.type is not None:
            schema = dict(type=self.type)
            del self.type
        elif hasattr(self, "oneOf") and self.oneOf:
            schema = dict(oneOf=self.oneOf)
            del self.oneOf
        else:
            schema = dict()
        if self.enum is not None:
            schema["enum"] = self.enum
            del self.enum
        if not hasattr(self, "oneOf") or self.oneOf is None:  # recreate for the null type
            self.oneOf = []
        if len(schema) > 0:  # i.e. the dict is populated
            self.oneOf.append(schema)
